from line_profiler import LineProfiler
import causaldag as cd
from causaldag.inference.structural import igsp
from causaldag.utils.ci_tests import gauss_ci_test, hsic_invariance_test
import numpy as np
import random
np.random.seed(1729)
random.seed(1729)

nnodes = 15
g = cd.rand.rand_weights(cd.rand.directed_erdos(nnodes, 3/(nnodes-1), 1))
iv_node = random
nsamples = 20
samples = {
    frozenset(): g.sample(nsamples),
    frozenset({iv_node}): g.sample_interventional_perfect({iv_node: cd.GaussIntervention(1, .1)}, nsamples)
}
corr = np.corrcoef(samples[frozenset()], rowvar=False)
suffstat = dict(C=corr, n=nsamples)
profiler = LineProfiler()


def run_igsp():
    for i in range(100):
        igsp(samples, suffstat, nnodes, gauss_ci_test, hsic_invariance_test, nruns=10)

profiler.add_function(igsp)
profiler.runcall(run_igsp)
profiler.print_stats()



from causaldag.classes import DAG
import numpy as np
from causaldag.utils import core_utils
from typing import Callable
from causaldag.classes.interventions import Intervention, SoftInterventionalDistribution, PerfectInterventionalDistribution


class SampleDAG(DAG):
    def __init__(self, nodes, arcs):
        super().__init__(set(nodes), arcs)
        self.conditionals = dict()
        self._node_list = list(nodes)
        self._node2ix = core_utils.ix_map_from_list(self._node_list)

    def set_conditional(self, node, conditional_distribution: Callable[[np.ndarray], np.ndarray]):
        self.conditionals[node] = conditional_distribution

    def sample(self, nsamples: int = 1) -> np.array:
        samples = np.zeros((nsamples, len(self._nodes)))
        t = self.topological_sort()
        for sample_num in range(nsamples):
            for node in t:
                node_ix = self._node2ix[node]
                parent_ixs = [self._node2ix[p] for p in self._parents[node]]
                samples[sample_num, node_ix] = self.conditionals[node](samples[sample_num, parent_ixs])
        return samples

    def sample_interventional(self, intervention: Intervention, nsamples: int = 1) -> np.ndarray:
        samples = np.zeros((nsamples, len(self._nodes)))

        t = self.topological_sort()
        for node in t:
            ix = self._node2ix[node]
            parent_ixs = [self._node2ix[p] for p in self._parents[node]]
            parent_vals = samples[:, parent_ixs]

            interventional_dist = intervention.get(node)
            if interventional_dist is not None:
                if isinstance(interventional_dist, SoftInterventionalDistribution):
                    samples[:, ix] = interventional_dist.sample(parent_vals, self, node)
                elif isinstance(interventional_dist, PerfectInterventionalDistribution):
                    samples[:, ix] = interventional_dist.sample(nsamples)
            else:
                for sample_num in range(nsamples):
                    samples[sample_num, ix] = self.conditionals[node](samples[sample_num, parent_ixs])

        return samples



# Author: Chandler Squires
"""
Base class for DAGs representing Gaussian distributions (i.e. linear SEMs with Gaussian noise).
"""
import operator as op
import itertools as itr
from typing import Any, Dict, Union, Set, Tuple, List

import numpy as np
from numpy import sqrt, diag
from numpy.linalg import inv
from scipy.linalg import ldl
from scipy.stats import norm

from causaldag.classes import DAG
from causaldag.classes.interventions import Intervention, SoftInterventionalDistribution, \
    PerfectInterventionalDistribution, PerfectIntervention, SoftIntervention, GaussIntervention, BinaryIntervention, \
    MultinomialIntervention, ConstantIntervention
from causaldag.utils import core_utils


class GaussDAG(DAG):
    def __init__(
            self,
            nodes: List,
            arcs: Union[Set[Tuple[Any, Any]], Dict[Tuple[Any, Any], float]],
            means=None,
            variances=None
    ):
        arcs_set = arcs if isinstance(arcs, set) else set(arcs.keys())
        self._weight_mat = np.zeros((len(nodes), len(nodes)))
        self._node_list = nodes
        self._node2ix = core_utils.ix_map_from_list(self._node_list)

        self._variances = np.ones(len(nodes)) if variances is None else np.array(variances, dtype=float)
        self._means = np.zeros((len(nodes))) if means is None else np.array(means)

        self._precision = None
        self._covariance = None
        self._correlation = None

        super().__init__(set(nodes), arcs_set)

        for node1, node2 in arcs:
            w = arcs[(node1, node2)] if isinstance(arcs, dict) else 1
            self._weight_mat[self._node2ix[node1], self._node2ix[node2]] = w

    def to_dag(self):
        """
        TODO

        Examples
        --------
        TODO
        """
        return DAG(nodes=set(self._node_list), arcs=self.arcs)

    @classmethod
    def from_amat(cls, weight_mat, nodes=None, means=None, variances=None):
        """
        Return a GaussDAG with arc weights specified by weight mat.

        Parameters
        ----------
        weight_mat:
            Matrix of edge weights, with weight[i, j] being the weight on the arc i->j.
        nodes
        means:
            Node residual means.
        variances:
            Node residual variances.

        Examples
        --------
        TODO
        """
        nodes = nodes if nodes is not None else list(range(weight_mat.shape[0]))
        arcs = {(i, j): w for (i, j), w in np.ndenumerate(weight_mat) if w != 0}
        return cls(nodes=nodes, arcs=arcs, means=means, variances=variances)

    @classmethod
    def from_covariance(cls, covariance_matrix: np.ndarray, node_order=None):
        """
        Return a GaussDAG with the specified covariance matrix and topological ordering of nodes.

        Parameters
        ----------
        covariance_matrix:
            The desired covariance matrix for the graph.
        node_order:
            The desired ordering of nodes.

        Examples
        --------
        TODO
        """
        if not core_utils.is_symmetric(covariance_matrix):
            raise ValueError('Covariance matrix is not symmetric')
        precision = inv(covariance_matrix)
        return GaussDAG.from_precision(precision, node_order)

    @classmethod
    def from_precision(cls, precision_matrix, node_order=None):
        """
        Return a GaussDAG with the specified precision matrix and topological ordering of nodes.

        Parameters
        ----------
        precision_matrix:
            The desired precision matrix for the graph.
        node_order:
            The desired ordering of nodes.

        Examples
        --------
        TODO
        """
        if not core_utils.is_symmetric(precision_matrix):
            raise ValueError('Covariance matrix is not symmetric')

        p = precision_matrix.shape[0]
        if node_order is None:
            node_order = list(range(p))

        # === permute precision matrix into  correct order for LDL
        precision_matrix = precision_matrix.copy()
        precision_matrix = precision_matrix[node_order]
        precision_matrix = precision_matrix[:, node_order]

        # === perform ldl decomposition and correct for floating point errors
        u, d, perm_ = ldl(precision_matrix, lower=False)
        u[np.isclose(u, 0)] = 0

        # === permute back
        inv_node_order = [i for i, j in sorted(enumerate(node_order), key=op.itemgetter(1))]
        u = u.copy()
        u = u[inv_node_order]
        u = u[:, inv_node_order]
        d = d.copy()
        d = d[inv_node_order]
        d = d[:, inv_node_order]

        amat = np.eye(p) - u
        variances = np.diag(d) ** -1

        # adj_mat[np.isclose(adj_mat, 0)] = 0
        return GaussDAG.from_amat(amat, variances=variances)

    def normalize_weights(self):
        """
        Return a GaussDAG with the same skeleton and correlation matrix, normalized so that each variable has unit
        variance.

        Examples
        --------
        TODO
        """
        normalized_weight_mat = np.zeros(self.weight_mat.shape)
        node_sds = sqrt(diag(self.covariance))
        num_nodes = self.nnodes
        for i, j in itr.combinations(range(num_nodes), 2):
            if self.weight_mat[i, j] != 0:
                normalized_weight_mat[i, j] = B[i, j] * node_sds[i] / node_sds[j]
        # TODO: variance
        return cd.GaussDAG.from_amat(normalized_weight_mat)

    def set_arc_weight(self, i, j, val):
        """
        Change the weight of the arc i->j to val

        Parameters
        ----------
        i:
            source node of arc.
        j:
            target node of arc.
        val:
            weight of the arc `i->j`.

        Examples
        --------
        TODO
        """
        self._weight_mat[self._node2ix[i], self._node2ix[j]] = val
        if val == 0 and (i, j) in self._arcs:
            super().remove_arc(i, j)
        if val != 0 and (i, j) not in self._arcs:
            super().add_arc(i, j)

    def set_node_variance(self, i, var):
        """
        Change the variance of node i to var

        Parameters
        ----------
        i:
            node whose variance to change.
        var:
            new variance.

        Examples
        --------
        TODO
        """
        self._variances[i] = var

    @property
    def nodes(self):
        return list(self._nodes)

    @property
    def arc_weights(self):
        return {(i, j): self._weight_mat[i, j] for i, j in self._arcs}

    @property
    def weight_mat(self):
        return self._weight_mat.copy()

    @property
    def variances(self):
        return self._variances.copy()

    @property
    def precision(self):
        self._ensure_precision()
        return self._precision.copy()

    @property
    def covariance(self):
        self._ensure_covariance()
        return self._covariance.copy()

    def _ensure_correlation(self):
        if self._correlation is None:
            S = self.covariance
            self._correlation = S / sqrt(diag(S)) / sqrt(diag(S)).reshape([-1, 1])

    @property
    def correlation(self):
        self._ensure_correlation()
        return self._correlation

    def partial_correlation(self, i, j, cond_set):
        """
        Return the partial correlation of i and j conditioned on `cond_set`.

        Parameters
        ----------
        i: first node.
        j: second node.
        cond_set: conditioning set.

        Examples
        --------
        TODO
        """
        cond_set = core_utils.to_set(cond_set)
        if len(cond_set) == 0:
            return self.correlation[i, j]
        else:
            theta = inv(self.correlation[np.ix_([i, j, *cond_set], [i, j, *cond_set])])
            return -theta[0, 1] / np.sqrt(theta[0, 0] * theta[1, 1])

    def add_arc(self, i, j, unsafe=False):
        """
        Add an arc to the graph with weight 1.

        Parameters
        ----------
        i:
        j:
        unsafe:

        Examples
        --------
        TODO
        """
        self.set_arc_weight(i, j, 1)

    def remove_arc(self, i, j, ignore_error=False):
        """
        Remove an arc from the graph

        Parameters
        ----------
        i:
        j:
        ignore_error:

        Examples
        --------
        TODO
        """
        self.set_arc_weight(i, j, 0)

    def add_node(self, node):
        """
        Add a node to the graph

        Parameters
        ----------
        node

        Examples
        --------
        TODO
        """
        self._node_list.append(node)
        self._weight_mat = np.zeros((len(self._node_list), len(self._node_list)))
        self._weight_mat[:-1, :-1] = None

    def remove_node(self, node, ignore_error=False):
        """
        Remove a node from the graph.

        Parameters
        ----------
        node:
        ignore_error:

        Examples
        --------
        TODO
        """
        del self._node_list[self._node2ix[node]]
        self._weight_mat = self._weight_mat[np.ix_(self._node_list, self._node_list)]
        super().remove_node(node)

    def add_arcs_from(self, arcs, unsafe=False):
        """
        TODO

        Parameters
        ----------
        arcs:
        unsafe:

        Examples
        --------
        TODO
        """
        for i, j in arcs:
            self.add_arc(i, j, unsafe=unsafe)

    def add_nodes_from(self, nodes):
        raise NotImplementedError
        pass

    def reverse_arc(self, i, j, ignore_error=False, unsafe=False):
        raise NotImplementedError
        pass

    def arcs_in_vstructures(self):
        return super().arcs_in_vstructures()

    def reversible_arcs(self):
        return super().reversible_arcs()

    def topological_sort(self):
        return super().topological_sort()

    def shd(self, other):
        return super().shd(other)

    def downstream(self, node):
        return super().downstream(node)

    def upstream(self, node):
        return super().upstream(node)

    def incident_arcs(self, node):
        return super().incident_arcs(node)

    def incoming_arcs(self, node):
        return super().incoming_arcs(node)

    def outgoing_arcs(self, node):
        return super().outgoing_arcs(node)

    def outdegree(self, node):
        return super().outdegree(node)

    def indegree(self, node):
        return super().indegree(node)

    def save_gml(self, filename):
        raise NotImplementedError

    def to_amat(self):
        """
        TODO

        Parameters
        ----------
        TODO

        Examples
        --------
        TODO
        """
        return self.weight_mat

    def cpdag(self):
        raise NotImplementedError

    def optimal_intervention_greedy(self, cpdag=None):
        return super().optimal_intervention_greedy(cpdag=cpdag)

    def backdoor(self, i, j):
        return super().backdoor(i, j)

    def frontdoor(self, i, j):
        return super().frontdoor(i, j)

    def dsep(self, i, j, C=None, verbose=False, certify=False):
        return super().dsep(i, j, C=C, verbose=verbose, certify=certify)

    def _ensure_precision(self):
        if self._precision is None:
            id_ = np.eye(len(self._nodes))
            a = self._weight_mat
            if (self._variances == 1).all():
                self._precision = (id_ - a) @ (id_ - a).T
            else:
                self._precision = (id_ - a) @ np.diag(self._variances ** -1) @ (id_ - a).T

    def _ensure_covariance(self):
        if self._covariance is None:
            id_ = np.eye(len(self._nodes))
            a = self._weight_mat
            id_min_a_inv = inv(id_ - a)
            if (self._variances == 1).all():
                self._covariance = id_min_a_inv.T @ id_min_a_inv
            else:
                self._covariance = id_min_a_inv.T @ np.diag(self._variances) @ id_min_a_inv

    def sample(self, nsamples: int = 1) -> np.array:
        """
        Return `nsamples` samples from the graph.

        Parameters
        ----------
        nsamples:
            Number of samples.

        Returns
        -------
        (nsamples x nnodes) matrix of samples.

        Examples
        --------
        TODO
        """
        samples = np.zeros((nsamples, len(self._nodes)))
        noise = np.zeros((nsamples, len(self._nodes)))
        for ix, (mean, var) in enumerate(zip(self._means, self._variances)):
            noise[:, ix] = np.random.normal(loc=mean, scale=var ** .5, size=nsamples)
        t = self.topological_sort()
        for node in t:
            ix = self._node2ix[node]
            parents = self._parents[node]
            if len(parents) != 0:
                parent_ixs = [self._node2ix[p] for p in self._parents[node]]
                parent_vals = samples[:, parent_ixs]
                samples[:, ix] = np.sum(parent_vals * self._weight_mat[parent_ixs, node], axis=1) + noise[:, ix]
            else:
                samples[:, ix] = noise[:, ix]
        return samples

    def sample_interventional_perfect(self, interventions: PerfectIntervention, nsamples: int = 1) -> np.array:
        """
        Return `nsamples` samples from the graph under a perfect intervention

        Parameters
        ----------
        interventions:
        nsamples:

        Returns
        -------
        (nsamples x nnodes) matrix of samples.
        """
        samples = np.zeros((nsamples, len(self._nodes)))
        noise = np.zeros((nsamples, len(self._nodes)))

        for ix, (node, mean, var) in enumerate(zip(self._node_list, self._means, self._variances)):
            interventional_dist = interventions.get(node)
            if interventional_dist is not None:
                noise[:, ix] = interventional_dist.sample(nsamples)
            else:
                noise[:, ix] = np.random.normal(loc=mean, scale=var ** .5, size=nsamples)

        t = self.topological_sort()
        for node in t:
            ix = self._node2ix[node]
            parents = self._parents[node]
            if node not in interventions and len(parents) != 0:
                parent_ixs = [self._node2ix[p] for p in self._parents[node]]
                parent_vals = samples[:, parent_ixs]
                samples[:, ix] = np.sum(parent_vals * self._weight_mat[parent_ixs, node], axis=1) + noise[:, ix]
            else:
                samples[:, ix] = noise[:, ix]

        return samples

    def sample_interventional_soft(self, intervention: SoftIntervention, nsamples: int = 1) -> np.array:
        """
        Return samples from the graph under a soft intervention
        """
        samples = np.zeros((nsamples, len(self._nodes)))
        noise = np.zeros((nsamples, len(self._nodes)))
        for ix, var in enumerate(self._variances):
            noise[:, ix] = np.random.normal(scale=var ** .5, size=nsamples)

        t = self.topological_sort()
        for node in t:
            ix = self._node2ix[node]
            parent_ixs = [self._node2ix[p] for p in self._parents[node]]
            parent_vals = samples[:, parent_ixs]

            interventional_dist = intervention.get(node)
            if interventional_dist is not None:
                samples[:, ix] = interventional_dist.sample(parent_vals, self, node)
            elif len(parent_ixs) != 0:
                samples[:, ix] = np.sum(parent_vals * self._weight_mat[parent_ixs, node], axis=1) + noise[:, ix]
            else:
                samples[:, ix] = noise[:, ix]

        return samples

    def sample_interventional(self, intervention: Intervention, nsamples: int = 1) -> np.ndarray:
        """
        TODO

        Parameters
        ----------
        intervention:

        Examples
        --------
        TODO
        """
        samples = np.zeros((nsamples, len(self._nodes)))
        noise = np.random.normal(size=[nsamples, len(self._nodes)])
        noise = noise * np.array(self._variances) ** .5 + self._means

        t = self.topological_sort()
        for node in t:
            ix = self._node2ix[node]
            parent_ixs = [self._node2ix[p] for p in self._parents[node]]
            parent_vals = samples[:, parent_ixs]

            interventional_dist = intervention.get(node)
            if interventional_dist is not None:
                if isinstance(interventional_dist, SoftInterventionalDistribution):
                    samples[:, ix] = interventional_dist.sample(parent_vals, self, node)
                elif isinstance(interventional_dist, PerfectInterventionalDistribution):
                    samples[:, ix] = interventional_dist.sample(nsamples)
            elif len(parent_ixs) != 0:
                samples[:, ix] = np.sum(parent_vals * self._weight_mat[parent_ixs, node], axis=1) + noise[:, ix]
            else:
                samples[:, ix] = noise[:, ix]

        return samples

    def interventional_covariance(self, intervened_nodes: set):
        remaining_nodes = [node for node in self._nodes if node not in intervened_nodes]

        id_ = np.eye(len(self._nodes))
        a = self._weight_mat
        a = a[np.ix_(remaining_nodes, remaining_nodes)]
        id_min_a_inv = np.linalg.inv(id_ - a)
        if (self._variances == 1).all():
            return id_min_a_inv.T @ id_min_a_inv
        else:
            return id_min_a_inv.T @ np.diag(self._variances) @ id_min_a_inv

    # def logpdf(self, samples: np.array, interventions: Intervention = None) -> np.array:
    #     self._ensure_covariance()
    #
    #     if interventions is None:
    #         return multivariate_normal.logpdf(samples, mean=self._means, cov=self._covariance)
    #     else:
    #         intervened_nodes = set(interventions.keys())
    #         remaining_nodes = [node for node in self._nodes if node not in intervened_nodes]
    #         samples = samples[:, remaining_nodes]
    #         adjusted_means = None
    #         adjusted_cov = self.interventional_covariance(intervened_nodes)
    #         return multivariate_normal.logpdf(samples, meabn=adjusted_means, cov=adjusted_cov)

    def logpdf(self, samples: np.array, interventions: PerfectIntervention = None,
               exclude_intervention_prob=True) -> np.array:
        # TODO this is about 10x slower than using multivariate_normal.logpdf with the covariance matrix
        # TODO can I speed this up? where is the time spent?

        sorted_nodes = self.topological_sort()
        nsamples = samples.shape[0]
        log_probs = np.zeros(nsamples)

        if interventions is None:
            for node in sorted_nodes:
                node_ix = self._node2ix[node]
                parent_ixs = [self._node2ix[p] for p in self._parents[node]]
                if len(parent_ixs) != 0:
                    parent_vals = samples[:, parent_ixs]
                    correction = (parent_vals * self._weight_mat[parent_ixs, node]).sum(axis=1)
                else:
                    correction = 0
                log_probs += norm.logpdf(samples[:, node_ix] - correction, scale=self._variances[node_ix] ** .5)
        else:
            for node in sorted_nodes:
                node_ix = self._node2ix[node]
                iv = interventions.get(node)
                if iv is not None:
                    if not exclude_intervention_prob:
                        if isinstance(iv, GaussIntervention):
                            log_probs += iv.logpdf(samples[:, node_ix])
                        else:
                            log_probs += np.log(iv.pdf(samples[:, node_ix]))
                else:
                    parent_ixs = [self._node2ix[p] for p in self._parents[node]]
                    parent_vals = samples[:, parent_ixs]
                    correction = (parent_vals * self._weight_mat[parent_ixs, node]).sum(axis=1)
                    log_probs += norm.logpdf(samples[:, node_ix] - correction, scale=self._variances[node_ix] ** .5)

        return log_probs


if __name__ == '__main__':
    iv = MultinomialIntervention(
        interventions=[
            ConstantIntervention(val=-1),
            ConstantIntervention(val=1),
            GaussIntervention(mean=2, variance=1),
        ],
        pvals=[.4, .4, .2]
    )

    iv = BinaryIntervention(
        intervention1=ConstantIntervention(val=-1),
        intervention2=ConstantIntervention(val=1)
    )
    import causaldag as cd

    B = np.zeros((3, 3))
    B[0, 1] = 1
    B[0, 2] = -1
    B[1, 2] = 4
    gdag = cd.GaussDAG.from_amat(B)
    iv = causaldag.classes.interventions.GaussIntervention(mean=0, variance=.1)
    gdag.sample_interventional_perfect({0: iv}, nsamples=100)

    s = gdag.sample(1000)
    # print(gdag.arcs)
    print(s.T @ s / 1000)
    print(gdag.covariance)
    s2 = gdag.sample_interventional_perfect({1: iv}, 1000)
    print(s2.T @ s2 / 1000)

    import matplotlib.pyplot as plt

    plt.clf()
    plt.ion()
    plt.scatter(s2[:, 1], s2[:, 2])
    plt.show()



from collections import defaultdict
from causaldag.utils import core_utils
import itertools as itr
import numpy as np
import random
from typing import List, Iterable, Set, Dict, Hashable, Tuple, FrozenSet, Union
from causaldag.classes.custom_types import Node, DirectedEdge, BidirectedEdge, UndirectedEdge


class CycleError(Exception):
    def __init__(self):
        super().__init__()
        # def __init__(self, source, target):
        #     self.source = source
        #     self.target = target
        #     message = '%s -> %s will cause a cycle' % (source, target)
        #     super().__init__(message)


class SpouseError(Exception):
    def __init__(self):
        super().__init__()
        # def __init__(self, ancestor, desc):
        #     self.ancestor = ancestor
        #     self.desc = desc
        #     message = '%s <-> %s cannot be added since %s is an ancestor of %s' % (ancestor, desc, ancestor, desc)
        #     super().__init__(message)


class AdjacentError(Exception):
    def __init__(self, node1, node2, arrow_type):
        self.node1 = node1
        self.node2 = node2
        self.arrow_type = arrow_type
        message = '%s %s %s cannot be added since %s and %s are already adjacent' % (
            node1, arrow_type, node2, node1, node2)
        super().__init__(message)


class NeighborError(Exception):
    def __init__(self, node, neighbors=None, parents=None, spouses=None):
        self.node = node
        self.neighbors = neighbors
        self.parents = parents
        self.spouses = spouses
        if self.neighbors:
            message = 'The node %s has neighbors %s. Nodes cannot have neighbors and parents/spouses.' % (
                node, ','.join(map(str, neighbors)))
        elif self.parents:
            message = 'The node %s has parents %s. Nodes cannot have neighbors and parents/spouses.' % (
                node, ','.join(map(str, parents)))
        elif self.spouses:
            message = 'The node %s has spouses %s. Nodes cannot have neighbors and parents/spouses.' % (
                node, ','.join(map(str, spouses)))
        super().__init__(message)


def path2str(path):
    return '->'.join(map(str, path))


class AncestralGraph:
    """
    Base class for ancestral graphs, used to represent causal models with latent variables.
    """

    def __init__(self, nodes=set(), directed=set(), bidirected=set(), undirected=set()):
        self._nodes = nodes.copy()
        self._directed = set()
        self._bidirected = set()
        self._undirected = set()

        self._neighbors = defaultdict(set)
        self._spouses = defaultdict(set)
        self._parents = defaultdict(set)
        self._children = defaultdict(set)
        self._adjacent = defaultdict(set)

        for i, j in directed:
            self._add_directed(i, j)
        for i, j in bidirected:
            self._add_bidirected(i, j)
        for i, j in undirected:
            self._add_undirected(i, j)

    def __eq__(self, other):
        if not isinstance(other, AncestralGraph):
            return False

        same_nodes = self._nodes == other._nodes
        same_directed = self._directed == other._directed
        same_bidirected = self._bidirected == other._bidirected
        same_undirected = self._undirected == other._undirected
        return same_nodes and same_directed and same_bidirected and same_undirected

    def copy(self):
        """
        Return a copy of the current ancestral graph.
        """
        return AncestralGraph(self.nodes, self.directed, self.bidirected, self.undirected)

    def induced_subgraph(self, nodes: Set[Node]):
        """
        Return the induced subgraph over only `nodes`

        Parameters
        ----------
        nodes:
            Set of nodes for the induced subgraph.

        Returns
        -------
        g:
            Induced subgraph over `nodes`.

        Examples
        --------
        >>> g = cd.AncestralGraph(bidirected={(1, 2), (1, 4)}, directed={(1, 3), (2, 3)})
        >>> g.induced_subgraph()
        TODO
        """
        new_directed = {(i, j) for i, j in self._directed if i in nodes and j in nodes}
        new_bidirected = {(i, j) for i, j in self._bidirected if i in nodes and j in nodes}
        new_undirected = {(i, j) for i, j in self._undirected if i in nodes and j in nodes}

        return AncestralGraph(nodes, directed=new_directed, bidirected=new_bidirected, undirected=new_undirected)

    def __str__(self):
        return 'Directed edges: %s, Bidirected edges: %s, Undirected edges: %s' % (
        self._directed, self._bidirected, self._undirected)

    def __repr__(self):
        return str(self)

    # === MUTATORS
    def add_node(self, node: Node):
        """
        Add a node to the ancestral graph.

        Parameters
        ----------
        node:
            a hashable Python object

        See Also
        --------
        add_nodes_from

        Examples
        --------
        >>> g = cd.AncestralGraph()
        >>> g.add_node(1)
        >>> g.add_node(2)
        >>> len(g.nodes)
        2
        """
        self._nodes.add(node)

    def add_nodes_from(self, nodes: Iterable[Node]):
        """
        Add a node to the ancestral graph.

        Parameters
        ----------
        nodes:
            an iterable of hashable Python objects

        See Also
        --------
        add_node

        Examples
        --------
        >>> g = cd.AncestralGraph()
        >>> g.add_nodes_from({1, 2})
        >>> len(g.nodes)
        2
        """
        for node in nodes:
            self._nodes.add(node)

    def _check_ancestral(self):
        self.topological_sort()

    def _mark_children_visited(self, node, any_visited, curr_path_visited, curr_path, stack):
        any_visited[node] = True
        curr_path_visited[node] = True
        curr_path.append(node)
        for child in self._children[node]:
            if not any_visited[child]:
                self._mark_children_visited(child, any_visited, curr_path_visited, curr_path, stack)
            elif curr_path_visited[child]:
                cycle = curr_path + [child]
                raise CycleError
        for spouse in self._spouses[node]:
            if curr_path_visited[spouse]:
                raise SpouseError
        curr_path.pop()
        curr_path_visited[node] = False
        stack.append(node)

    def topological_sort(self) -> list:
        """
        Return a linear order that is consistent with the partial order implied by ancestral relations of this graph.

        Examples
        --------
        TODO
        """
        any_visited = {node: False for node in self._nodes}
        curr_path_visited = {node: False for node in self._nodes}
        curr_path = []
        stack = []
        for node in self._nodes:
            if not any_visited[node]:
                self._mark_children_visited(node, any_visited, curr_path_visited, curr_path, stack)
        return list(reversed(stack))

    def add_directed(self, i: Node, j: Node):
        """
        Add a directed edge from node `i` to node `j`.

        Parameters
        ----------
        i:
            source of directed edge.
        j:
            target of directed edge.

        Examples
        --------
        TODO
        """
        self._add_directed(i, j)
        try:
            self._check_ancestral()
        except CycleError as e:
            self.remove_directed(i, j)
            raise e

    def add_bidirected(self, i: Node, j: Node):
        """
        Add a bidirected edge between nodes `i` and `j`.

        Parameters
        ----------
        i:
            first endpoint of bidirected edge.
        j:
            second endpoint of bidirected edge.

        Examples
        --------
        TODO
        """
        self._add_bidirected(i, j)
        try:
            self._add_bidirected(i, j)
        except CycleError as e:
            self.remove_bidirected(i, j)
            raise e

    def add_undirected(self, i: Node, j: Node):
        """
        Add an undirected edge between nodes `i` and `j`.

        Parameters
        ----------
        i:
            first endpoint of undirected edge.
        j:
            second endpoint of undirected edge.

        Examples
        --------
        TODO
        """
        self._add_undirected(i, j)

    def _add_directed(self, i: Node, j: Node, ignore_error=False):
        if self.has_directed(i, j):
            return

        # === CHECK REMAINS ANCESTRAL
        if not ignore_error and self._neighbors[j]:
            raise NeighborError(j, self._neighbors[j])

        # === CHECK i AND j NOT ALREADY ADJACENT
        if i in self._adjacent[j]:
            if ignore_error:
                if self.has_directed(j, i):
                    self.remove_directed(j, i)
                elif self.has_bidirected(i, j):
                    self.remove_bidirected(i, j)
                else:
                    self.remove_undirected(i, j)
            else:
                raise AdjacentError(i, j, '->')

        self._nodes.add(i)
        self._nodes.add(j)
        self._directed.add((i, j))
        self._parents[j].add(i)
        self._children[i].add(j)

        self._adjacent[i].add(j)
        self._adjacent[j].add(i)

    def _add_bidirected(self, i: Node, j: Node, ignore_error=False):
        if self.has_bidirected(i, j):
            return

        # === CHECK REMAINS ANCESTRAL
        if not ignore_error and self._neighbors[i]:
            raise NeighborError(i, neighbors=self._neighbors[i])
        if not ignore_error and self._neighbors[j]:
            raise NeighborError(j, neighbors=self._neighbors[j])

        # === CHECK i AND j NOT ALREADY ADJACENT
        if i in self._adjacent[j]:
            if ignore_error:
                if self.has_directed(i, j):
                    self.remove_directed(i, j)
                elif self.has_directed(j, i):
                    self.remove_directed(j, i)
                else:
                    self.remove_undirected(i, j)
            else:
                raise AdjacentError(i, j, '<->')

        self._nodes.add(i)
        self._nodes.add(j)
        self._bidirected.add(frozenset({i, j}))
        self._spouses[j].add(i)
        self._spouses[i].add(j)

        self._adjacent[i].add(j)
        self._adjacent[j].add(i)

    def _add_undirected(self, i: Node, j: Node, ignore_error=False):
        if self.has_undirected(i, j):
            return

        # === CHECK REMAINS ANCESTRAL
        if self._parents[i]:
            raise NeighborError(i, parents=self._parents[i])
        if self._spouses[i]:
            raise NeighborError(i, spouses=self._spouses[i])
        if self._parents[j]:
            raise NeighborError(j, parents=self._parents[j])
        if self._spouses[j]:
            raise NeighborError(j, spouses=self._spouses[j])

        # === CHECK i AND j NOT ALREADY ADJACENT
        if i in self._adjacent[j]:
            if ignore_error:
                if self.has_directed(i, j):
                    self.remove_directed(i, j)
                elif self.has_directed(j, i):
                    self.remove_directed(j, i)
                else:
                    self.remove_bidirected(i, j)
            else:
                raise AdjacentError(i, j, '-')

        self._nodes.add(i)
        self._nodes.add(j)
        self._undirected.add(frozenset({i, j}))
        self._neighbors[j].add(i)
        self._neighbors[i].add(j)

        self._adjacent[i].add(j)
        self._adjacent[j].add(i)

    def remove_node(self, node: Node, ignore_error=False):
        """
        Remove `node`.

        Parameters
        ----------
        node

        Examples
        --------
        TODO
        """
        try:
            self._nodes.remove(node)
            for parent in self._parents[node]:
                self._children[parent].remove(node)
                self._adjacent[parent].remove(node)
                self._directed.remove((parent, node))
            for child in self._children[node]:
                self._parents[child].remove(node)
                self._adjacent[child].remove(node)
                self._directed.remove((node, child))
            for spouse in self._spouses[node]:
                self._spouses[spouse].remove(node)
                self._adjacent[spouse].remove(node)
                self._bidirected.remove(frozenset({spouse, node}))
            for nbr in self._neighbors[node]:
                self._neighbors[nbr].remove(node)
                self._adjacent[nbr].remove(node)
                self._undirected.remove(frozenset({nbr, node}))

            del self._children[node]
            del self._parents[node]
            del self._spouses[node]
            del self._neighbors[node]
            del self._adjacent[node]
        except KeyError as e:
            if ignore_error:
                pass
            else:
                raise e

    def remove_directed(self, i: Node, j: Node, ignore_error=False):
        """
        Remove the directed edge from `i` to `j`.

        Parameters
        ----------
        i:
            source of directed edge.
        j:
            target of directed edge.

        Examples
        --------
        TODO
        """
        try:
            self._directed.remove((i, j))
            self._children[i].remove(j)
            self._parents[j].remove(i)
            self._adjacent[i].remove(j)
            self._adjacent[j].remove(i)
        except KeyError as e:
            if ignore_error:
                pass
            else:
                raise e

    def remove_bidirected(self, i: Node, j: Node, ignore_error=False):
        """
        Remove the bidirected edge between `i` and `j`.

        Parameters
        ----------
        i:
            first endpoint of bidirected edge.
        j:
            second endpoint of bidirected edge.

        Examples
        --------
        TODO
        """
        try:
            self._bidirected.remove(frozenset({i, j}))
            self._spouses[i].remove(j)
            self._spouses[j].remove(i)
            self._adjacent[i].remove(j)
            self._adjacent[j].remove(i)
        except KeyError as e:
            if ignore_error:
                pass
            else:
                raise e

    def remove_undirected(self, i: Node, j: Node, ignore_error=False):
        """
        Remove the undirected edge between `i` and `j`.

        Parameters
        ----------
        i:
            first endpoint of undirected edge.
        j:
            second endpoint of undirected edge.

        Examples
        --------
        TODO
        """
        try:
            self._undirected.remove(frozenset({i, j}))
            self._neighbors[i].remove(j)
            self._neighbors[j].remove(i)
            self._adjacent[i].remove(j)
            self._adjacent[j].remove(i)
        except KeyError as e:
            if ignore_error:
                pass
            else:
                raise e

    def remove_edge(self, i: Node, j: Node, ignore_error=False):
        """
        TODO

        Parameters
        ----------
        i:
            first endpoint of edge.
        j:
            second endpoint of edge.

        Examples
        --------
        TODO
        """
        if self.has_bidirected(i, j):
            self.remove_bidirected(i, j)
        elif self.has_directed(i, j):
            self.remove_directed(i, j)
        elif self.has_directed(j, i):
            self.remove_directed(j, i)
        elif self.has_undirected(i, j):
            self.remove_undirected(i, j)
        elif not ignore_error:
            raise KeyError

    def remove_edges(self, edges: Iterable):
        """
        TODO

        Parameters
        ----------
        edges
            TODO

        Examples
        --------
        TODO
        """
        for i, j in edges:
            self.remove_edge(i, j)

    # === PROPERTIES
    @property
    def nodes(self) -> Set[Node]:
        return self._nodes.copy()

    @property
    def nnodes(self) -> int:
        return len(self._nodes)

    @property
    def directed(self) -> Set[DirectedEdge]:
        return self._directed.copy()

    @property
    def num_directed(self) -> int:
        return len(self._directed)

    @property
    def bidirected(self) -> Set[BidirectedEdge]:
        return self._bidirected.copy()

    @property
    def num_bidirected(self) -> int:
        return len(self._bidirected)

    @property
    def undirected(self) -> Set[UndirectedEdge]:
        return self._undirected.copy()

    @property
    def num_undirected(self) -> int:
        return len(self._undirected)

    @property
    def num_edges(self) -> int:
        return self.num_directed + self.num_bidirected + self.num_undirected

    @property
    def skeleton(self) -> Set[UndirectedEdge]:
        return {frozenset({i, j}) for i, j in self._bidirected | self._undirected | self._directed}

    def children_of(self, i: Node) -> Set[Node]:
        return self._children[i].copy()

    def parents_of(self, i: Node) -> Set[Node]:
        return self._parents[i].copy()

    def spouses_of(self, i: Node) -> Set[Node]:
        return self._spouses[i].copy()

    def neighbors_of(self, i: Node) -> Set[Node]:
        return self._neighbors[i].copy()

    def _add_ancestors(self, ancestors, node, exclude_arcs=set()):
        for parent in self._parents[node]:
            if parent not in ancestors and (parent, node) not in exclude_arcs:
                ancestors.add(parent)
                self._add_ancestors(ancestors, parent, exclude_arcs=exclude_arcs)

    def _add_descendants(self, descendants, node, exclude_arcs=set()):
        for child in self._children[node]:
            if child not in descendants and (child, node) not in exclude_arcs:
                descendants.add(child)
                self._add_descendants(descendants, child, exclude_arcs=exclude_arcs)

    def ancestors_of(self, nodes, exclude_arcs=set()) -> Set[Node]:
        """
        Return the nodes upstream of node

        Parameters
        ----------
        nodes:
            Set of nodes.
        exclude_arcs:
            TODO

        See Also
        --------
        descendants_of

        Return
        ------
        Set[node]
            Return all nodes j such that there is a directed path from j to node.

        Example
        -------
        """
        ancestors = set()
        if not isinstance(nodes, set):
            self._add_ancestors(ancestors, nodes, exclude_arcs=exclude_arcs)
        else:
            return set.union(*(self.ancestors_of(node) for node in nodes))
        return ancestors

    def ancestor_dict(self) -> dict:
        """
        Return a dictionary from each node to its ancestors.

        See Also
        --------
        ancestors_of

        Return
        ------
        Dict[node,Set]
            Mapping node to ancestors

        Example
        -------
        """
        top_sort = self.topological_sort()

        node2ancestors_plus_self = defaultdict(set)
        for node in top_sort:
            node2ancestors_plus_self[node].add(node)
            for child in self._children[node]:
                node2ancestors_plus_self[child].update(node2ancestors_plus_self[node])

        for node in self._nodes:
            node2ancestors_plus_self[node] -= {node}

        return core_utils.defdict2dict(node2ancestors_plus_self, self._nodes)

    def descendants_of(self, node: Node, exclude_arcs=set()) -> Set[Node]:
        """
        Return the nodes downstream of node

        Parameters
        ----------
        node:
            The node.

        See Also
        --------
        ancestors_of

        Return
        ------
        Set[node]
            Return all nodes j such that there is a directed path from node j.

        Example
        -------
        """
        descendants = set()
        self._add_descendants(descendants, node, exclude_arcs=exclude_arcs)
        return descendants

    def has_directed(self, i: Node, j: Node) -> bool:
        """
        Check if this graph has the directed edge i->j.

        See Also
        --------
        has_bidirected
        has_undirected
        has_any_edge

        Parameters
        ----------
        i:
            Node.
        j:
            Node.

        Examples
        --------
        TODO
        """
        return (i, j) in self._directed

    def has_bidirected(self, i: Node, j: Node) -> bool:
        """
        Check if this graph has a bidirected edge between `i` and `j`.

        See Also
        --------
        has_directed
        has_undirected
        has_any_edge

        Parameters
        ----------
        i:
            Node.
        j:
            Node.

        Examples
        --------
        TODO
        """
        return frozenset({i, j}) in self._bidirected

    def has_undirected(self, i: Node, j: Node) -> bool:
        """
        Check if this graph has an undirected edge between `i` and `j`.

        See Also
        --------
        has_directed
        has_bidirected
        has_any_edge

        Parameters
        ----------
        i:
            Node.
        j:
            Node.

        Examples
        --------
        TODO
        """
        return frozenset({i, j}) in self._undirected

    def has_any_edge(self, i: Node, j: Node) -> bool:
        """
        Check if i and j are adjacent in this graph.

        See Also
        --------
        has_directed
        has_bidirected
        has_undirected

        Parameters
        ----------
        i:
            Node.
        j:
            Node.

        Examples
        --------
        TODO
        """
        return self.has_directed(i, j) or self.has_directed(j, i) or self.has_bidirected(i, j) or self.has_undirected(i,
                                                                                                                      j)

    def vstructures(self) -> Set[Tuple]:
        """
        TODO

        Examples
        --------
        TODO
        """
        vstructs = set()
        for node in self._nodes:
            for p1, p2 in itr.combinations(self._parents[node] | self._spouses[node], 2):
                if not self.has_any_edge(p1, p2):
                    p1_, p2_ = sorted((p1, p2))
                    vstructs.add((p1_, node, p2_))
        return vstructs

    def colliders(self) -> set:
        """
        TODO

        Examples
        --------
        TODO
        """
        return {node for node in self._nodes if len(self._parents[node] | self._spouses[node]) >= 2}

    def _bidirected_reachable(self, node, tmp: Set[Node], visited: Set[Node]) -> Set[Node]:
        visited.add(node)
        tmp.add(node)
        for spouse in filter(lambda spouse: spouse not in visited, self._spouses[node]):
            tmp = self._bidirected_reachable(spouse, tmp, visited)
        return tmp

    def c_components(self) -> List[set]:
        """
        Return the c-components of this graph.

        Return
        ------
        List[Set[node]]
            Return the partition of nodes coming from the relation of reachability by bidirected edges.

        Examples
        --------
        TODO
        """
        node_queue = self._nodes.copy()
        components = []
        visited_nodes = set()

        while node_queue:
            node = node_queue.pop()
            if node not in visited_nodes:
                components.append(self._bidirected_reachable(node, set(), visited_nodes))

        return components

    def district_of(self, node: Node) -> Set[Node]:
        """
        Return the district of a node, i.e., the set of nodes reachable by bidirected edges.

        Return
        ------
        Set[node]
            The district of node.

        Examples
        --------
        TODO
        """
        return self._bidirected_reachable(node, set(), set())

    def discriminating_paths(self, verbose=False) -> Dict[Tuple, str]:
        """
        TODO

        Parameters
        ----------
        TODO

        Examples
        --------
        TODO
        """
        colliders = self.colliders()
        discriminating_paths = {}
        if verbose: print("Checking discriminating paths")
        for j, parents in self._parents.items():  # potential endpoints of discriminating paths
            if verbose: print(j)
            if not parents:
                continue
            nonadjacent = self._nodes - parents - self._children[j] - self._spouses[j] - {j}
            if verbose: print(f"Checking node {j} and non-adjacent nodes {nonadjacent}")
            for i in nonadjacent:  # potential start points of discriminating paths
                # search all paths that satisfy discriminating path criteria
                path_queue = [
                    [i, k]
                    for k in self._spouses[i] | self._children[i]
                    if k in colliders and j in self._children[k]
                ]
                while path_queue:
                    path = path_queue.pop(0)
                    final_node = path[-1]
                    # check if path is discriminating for the next node
                    for k in filter(lambda k: k not in path, self._spouses[final_node]):
                        if j in self._spouses[k]:
                            full_path = path.copy()
                            full_path.extend([k, j])
                            discriminating_paths[tuple(full_path)] = 'c'
                        elif j in self._children[k]:
                            full_path = path.copy()
                            full_path.extend([k, j])
                            discriminating_paths[tuple(full_path)] = 'n'
                    for k in filter(lambda k: k not in path, self._parents[final_node]):
                        if j in self._children[k]:
                            full_path = path.copy()
                            full_path.extend([k, j])
                            discriminating_paths[tuple(full_path)] = 'n'

                    # extend path
                    for k in self._spouses[final_node]:
                        if k not in path and k in colliders and j in self._children[k]:
                            new_path = path.copy()
                            new_path.append(k)
                            path_queue.append(new_path)
        return discriminating_paths

    def _reachable(self, start_node, end_node, visited=set(), allowed_edges={'b', 'u', 'c', 'p'},
                   predicate=lambda node: True, verbose=False):
        allowed_nbrs = set()
        if 'b' in allowed_edges:
            allowed_nbrs.update(self._spouses[start_node])
        if 'u' in allowed_edges:
            allowed_nbrs.update(self._neighbors[start_node])
        if 'c' in allowed_edges:
            allowed_nbrs.update(self._children[start_node])
        if 'p' in allowed_edges:
            allowed_nbrs.update(self._parents[start_node])

        allowed_nbrs = {nbr for nbr in allowed_nbrs if predicate(nbr)}
        if verbose: print(f"Allowed neighbors of {start_node}: {allowed_nbrs}")
        if verbose: print(f"Visited: {visited}")

        results = []
        for nbr in allowed_nbrs:
            if nbr in visited:
                continue
            visited.add(nbr)
            if nbr == end_node:
                if verbose: print("Reached end node")
                return True
            results.append(
                self._reachable(nbr, end_node, visited=visited, allowed_edges=allowed_edges, predicate=predicate,
                                verbose=verbose))

        if verbose: print("reachability results:", results)
        return any(results)

    # === ???
    def pairwise_markov_statements(self) -> Set[Tuple[Node, Node, FrozenSet[Node]]]:
        """
        TODO

        Examples
        --------
        TODO
        """
        statements = set()
        for i, j in itr.combinations(self._nodes, 2):
            if not self.has_any_edge(i, j):
                statements.add((i, j, frozenset(self.ancestors_of(i) | self.ancestors_of(j) - {i, j})))
        return statements

    def is_imap(self, other, certify: bool = False) -> bool:
        """
        Check if this graph is an IMAP of the `other` graph, i.e., all m-separation statements in this graph
        are also m-separation statements in `other`.

        Parameters
        ----------
        other:
            Another DAG.
        certify:
            TODO

        See Also
        --------
        is_minimal_imap

        Examples
        --------
        >>> g = cd.AncestralGraph(arcs={(1, 2), (3, 2)})
        TODO
        """
        if not self.is_maximal():
            raise Exception("Your graph is not maximal")

        certificate = next(
            ((i, j, S) for i, j, S in self.pairwise_markov_statements() if not other.msep(i, j, S)),
            None)
        is_imap_ = certificate is None
        if certify:
            return is_imap_, certificate
        else:
            return is_imap_

    # def is_minimal_imap(self, other, certify=False):
    #     print("THIS HAS NOT BEEN TESTED")
    #     certificate = next((
    #         i, j for i, j in self._directed | self._bidirected
    #         if other.msep(i, j, self.ancestors_of(i) | self.ancestors_of(j) - {i, j})
    #     ), False)
    #     res = not certificate and self.is_imap(other)
    #     if not certify:
    #         return res
    #     else:
    #         return res, certificate

    def is_minimal_imap(self, other, certify: bool = False, check_imap=True) -> bool:
        """
        TODO

        Parameters
        ----------
        TODO

        Examples
        --------
        TODO
        """
        if check_imap and not self.is_imap(other):
            return False, None

        for i, j in random.sample(list(self._directed) + list(self._bidirected),
                                  self.num_bidirected + self.num_directed):
            new_mag = self.copy()
            if self.has_bidirected(i, j):
                new_mag.remove_bidirected(i, j)
            if self.has_directed(i, j):
                new_mag.remove_directed(i, j)
            if new_mag.is_maximal() and new_mag.is_imap(other):
                if certify:
                    return False, (i, j)
                else:
                    return False
        if certify:
            return True, None
        else:
            return True

    def is_minimal_imap2(self, other, certify=False, check_imap=True, validate=False):
        if check_imap and not self.is_imap(other):
            return False, None

        for i, j in random.sample(list(self._directed) + list(self._bidirected),
                                  self.num_directed + self.num_bidirected):
            if other.msep(i, j, self.ancestors_of(i) | self.ancestors_of(j) - {i, j}):
                new_mag = self.copy()
                if self.has_bidirected(i, j):
                    new_mag.remove_bidirected(i, j)
                else:
                    new_mag.remove_directed(i, j)
                if new_mag.is_maximal():
                    if validate:
                        if not new_mag.is_imap(other):
                            raise Exception
                    if certify:
                        return False, (i, j)
                    else:
                        return False
        if certify:
            return True, None
        else:
            return True

    def is_minimal_imap3(self, other, certify=False, check_imap=True, validate=False, verbose=False):
        if check_imap and not self.is_imap(other):
            return False, None

        for i, j in random.sample(list(self._directed) + list(self._bidirected),
                                  self.num_directed + self.num_bidirected):
            new_mag = self.copy()
            if self.has_bidirected(i, j):
                new_mag.remove_bidirected(i, j)
            else:
                new_mag.remove_directed(i, j)
            current_markov_blanket = set.union(*(set(v) for v in self.markov_blanket(j).values())) | self.district_of(j)
            new_markov_blanket = set.union(*(set(v) for v in new_mag.markov_blanket(j).values())) | new_mag.district_of(
                j)
            mb_difference = (current_markov_blanket - new_markov_blanket - {j}) | {i}
            rest = new_markov_blanket - {i, j}
            if verbose: print(f'i={i}, j={j}, mb_diff={mb_difference}, rest={rest}')
            if verbose: print("H", self)
            if verbose: print("G", other)

            if other.msep(i, mb_difference, rest) and new_mag.is_maximal():
                print('here')
                if validate:
                    if not new_mag.is_imap(other):
                        raise Exception
                if certify:
                    return False, (i, j)
                else:
                    return False
        if certify:
            return True, None
        else:
            return True

    def is_minimal_imap4(self, other, certify=False, check_imap=True, validate=False, extra_validate=False,
                         verbose=False):
        if check_imap and not self.is_imap(other):
            raise Exception("Not an IMAP")
            print("isn't imap")
            return False, None

        if extra_validate:
            for i, j in self._directed | self._bidirected:
                new_mag = self.copy()
                new_mag.remove_edge(i, j)
                s = new_mag.induced_subgraph(new_mag.ancestors_of({i, j}) | {i, j}).markov_blanket(j, flat=True) - {j}
                if other.msep(i, j, s) and new_mag.is_maximal():
                    if not new_mag.is_imap(other):
                        raise Exception("CI test not sufficient", new_mag, other, i, j, s)
            print('extra validated')

        for i, j in random.sample(list(self._directed) + list(self._bidirected),
                                  self.num_directed + self.num_bidirected):
            change = False
            new_mag = self.copy()
            new_mag.remove_edge(i, j)

            # works:
            set1 = (new_mag.markov_blanket(j, flat=True) & new_mag.ancestors_of(i)) | new_mag.parents_of(j)
            remove_edge = other.msep(i, j, set1)

            # new:
            set2 = new_mag.induced_subgraph(new_mag.ancestors_of({i, j}) | {i, j}).markov_blanket(j, flat=True) - {i, j}
            remove_edge2 = other.msep(i, j, set2)
            # print(i, j, set2)
            set3 = new_mag.markov_blanket(j, flat=True) & new_mag.ancestors_of({j}) - {j}

            # if set2 != set3:
            #     print(new_mag, j, set2, set3)

            if remove_edge2:
                change = True
            # if self.has_bidirected(i, j) and other.msep(i, j, self.parents_of(j)) and other.msep(i, j, self.parents_of(i)):
            #     new_mag = self.copy()
            #     new_mag.remove_bidirected(i, j)
            #     change = True
            # elif self.has_directed(i, j) and other.msep(i, j, self.parents_of(j) - {i}):
            #     new_mag = self.copy()
            #     new_mag.remove_directed(i, j)
            #     change = True
            if change and new_mag.is_maximal():
                if validate:
                    if not new_mag.is_imap(other):
                        raise Exception("CI test isn't sufficient: new MAG is not an IMAP")
                if certify:
                    return False, (i, j)
                else:
                    return False

        if certify:
            return True, None
        else:
            return True

    def markov_blanket(self, node, flat: bool = False) -> Union[Set[Node], Dict]:
        """
        Return the Markov blanket of a node with respect to the whole graph.

        Parameters
        ----------
        node: The node whose Markov blanket to find.
        flat: if True, return the Markov blanket as a set, otherwise return a dictionary mapping nodes in the district
            of node to their parents.

        Returns
        -------
        The Markov blanket of node, including the node itself.
        """
        if not flat:
            return {d: self._parents[d] for d in self.district_of(node)}
        else:
            district = self.district_of(node)
            return district | set.union(*(self._parents[d] for d in district)) | {node}

    def resolved_quasisinks(self, other):
        res_qsinks = set()
        while True:
            new_resolved = {
                node for node in self._nodes - res_qsinks
                if not (self._children[node] - res_qsinks) and
                                 not (other._children[node] - res_qsinks) and
                                 self.markov_blanket(node) == other.markov_blanket(node)
            }
            res_qsinks.update(new_resolved)
            if not new_resolved:
                break

        return res_qsinks

    def is_maximal(self, new=True, verbose=False) -> bool:
        """
        TODO

        Parameters
        ----------
        TODO

        Examples
        --------
        TODO
        """
        new_mag = self.copy()
        new_mag.to_maximal(new=new, verbose=verbose)
        return new_mag == self

    def to_maximal(self, new=True, verbose=False):
        """
        TODO

        Parameters
        ----------
        TODO

        Examples
        --------
        TODO
        """
        if new:
            converged = False
            while not converged:
                # === NEED DICTIONARY OF ANCESTORS AND C-COMPONENTS TO CHECK INDUCING PATHS
                ancestor_dict = self.ancestor_dict()
                c_components = self.c_components()
                node2component = dict()
                for ix, component in enumerate(c_components):
                    for node in component:
                        node2component[node] = ix
                if verbose: print('==========')
                if verbose: print('Ancestor dict:', ancestor_dict)
                if verbose: print('C components', c_components)

                # === FIND INDUCING PATHS BETWEEN PAIRS OF NODE
                induced_pairs = []

                non_adjacent_pairs = ((i, j) for i, j in itr.combinations(self._nodes, 2) if
                                      not self.has_any_edge(i, j))
                for node1, node2 in non_adjacent_pairs:
                    check_ancestry = lambda node: node in ancestor_dict[node1] or node in ancestor_dict[node2]
                    nbrs1 = self._children[node1] | self._spouses[node1]
                    nbrs2 = self._children[node2] | self._spouses[node2]
                    if verbose: print(f"-------------\nChecking {node1} and {node2}")

                    # ONLY CHECK PATHS BETWEEN SPOUSES/CHILDREN THAT ARE IN THE SAME C-COMPONENT
                    for nbr1, nbr2 in itr.product(nbrs1, nbrs2):
                        same_component = node2component[nbr1] == node2component[nbr2]
                        if same_component and nbr1 in ancestor_dict[node2] and nbr2 in ancestor_dict[node1]:
                            if verbose: print(f"Checking neighbors {nbr1} (for {node1}) and {nbr2} (for {node2})")
                            if self._reachable(nbr1, nbr2, visited=set(), allowed_edges={'b'}, predicate=check_ancestry,
                                               verbose=verbose):
                                if verbose: print("Reachable")
                                induced_pairs.append((node1, node2))
                                continue
                            elif verbose:
                                print("No path")
                if verbose: print(f"found induced pairs: {induced_pairs}")
                for node1, node2 in induced_pairs:
                    self.add_bidirected(node1, node2)

                converged = len(induced_pairs) == 0
                # print('converged:', converged)
        else:
            for i, j in itr.combinations(self._nodes, r=2):
                if not self.has_any_edge(i, j):
                    never_msep = not any(self.msep(i, j, S) for S in core_utils.powerset(self._nodes - {i, j}))
                    if never_msep: self.add_bidirected(i, j)

    def to_pag(self):
        raise NotImplementedError

    # === CONVERTERS
    def to_amat(self) -> np.ndarray:
        """
        TODO

        Examples
        --------
        TODO
        """
        amat = np.zeros([self.nnodes, self.nnodes])
        for i, j in self.directed:
            amat[i, j] = 2
            amat[j, i] = 3
        for i, j in self.bidirected:
            amat[i, j] = 2
            amat[j, i] = 2
        for i, j in self.undirected:
            amat[i, j] = 3
            amat[j, i] = 3
        return amat

    @staticmethod
    def from_amat(amat: np.ndarray):
        """
        TODO

        Parameters
        ----------
        amat

        Examples
        --------
        TODO
        """
        p = amat.shape[0]
        directed = set()
        bidirected = set()
        undirected = set()
        for i, j in itr.combinations(set(range(p)), 2):
            vij = amat[i, j]
            vji = amat[j, i]
            if vij == 2 and vji == 3:  # arrowhead at j
                directed.add((i, j))
            elif vij == 3 and vji == 2:  # arrowhead at i
                directed.add((j, i))
            elif vij == 2 and vji == 2:  # arrowheads at both
                bidirected.add((i, j))
            elif vij == 3 and vji == 3:  # no arrowhead
                undirected.add((i, j))
        return AncestralGraph(set(range(p)), directed, bidirected, undirected)

    # === COMPARISON
    def markov_equivalent(self, other) -> bool:
        """
        Check if this graph is Markov equivalent to the graph `other`. Two graphs are Markov equivalent iff.
        they have the same skeleton, same v-structures, and if whenever there is the same discriminating path for some
        node in both graphs, the node is a collider on that path in one graph iff. it is a collider on that path in
        the other graph.

        Parameters
        ----------
        other:
            Another AncestralGraph.

        Examples
        --------
        TODO
        """
        same_skeleton = self.skeleton == other.skeleton
        same_vstructures = self.vstructures() == other.vstructures()

        self_discriminating_paths = self.discriminating_paths()
        other_discriminating_paths = other.discriminating_paths()
        shared_disc_paths = set(self_discriminating_paths.keys()) & set(other_discriminating_paths)
        same_discriminating = all(
            self_discriminating_paths[path] == other_discriminating_paths[path]
            for path in shared_disc_paths
        )

        return same_skeleton and same_vstructures and same_discriminating

    def get_all_mec(self):
        """
        TODO

        Examples
        --------
        TODO
        """
        visited = set()
        queue = [self]
        mags = []

        while queue:
            mag = queue.pop()
            mags.append(mag)
            curr_dir, curr_bidir = frozenset(mag._directed), frozenset({frozenset({*e}) for e in mag._bidirected})
            visited.add((curr_dir, curr_bidir))
            lmcs_dir, lmcs_bidir = mag.legitimate_mark_changes()
            for i, j in lmcs_dir:
                new_dir = curr_dir - {(i, j)}
                new_bidir = curr_bidir | {frozenset({i, j})}
                if (new_dir, new_bidir) not in visited:
                    new_mag = mag.copy()
                    new_mag.remove_directed(i, j)
                    new_mag.add_bidirected(i, j)
                    queue.append(new_mag)
            for i, j in lmcs_bidir:
                new_dir = curr_dir | {(i, j)}
                new_bidir = curr_bidir - {frozenset({i, j})}
                if (new_dir, new_bidir) not in visited:
                    new_mag = mag.copy()
                    new_mag.remove_bidirected(i, j)
                    new_mag.add_directed(i, j)
                    queue.append(new_mag)

        return mags

    def shd_skeleton(self, other) -> int:
        """
        Compute the structure Hamming distance between the skeleton of this graph and the skeleton of another graph.

        Parameters
        ----------
        other:
            the graph to which the SHD of the skeleton will be computed.

        Return
        ------
        int
            The structural Hamming distance between :math:`G_1` and :math:`G_2` is the minimum number of arc additions,
            deletions, and reversals required to transform :math:`G_1` into :math:`G_2` (and vice versa).

        Example
        -------
        >>> TODO
        """
        return len(self.skeleton.symmetric_difference(other.skeleton))

    def as_hashed(self):
        """
        TODO

        Examples
        --------
        TODO
        """
        return frozenset(self._directed), frozenset(self._bidirected), frozenset(self._undirected)

    # === Algorithms
    def _add_upstream(self, upstream: set, node: Node):
        for parent in self._parents[node]:
            if parent not in upstream:
                upstream.add(parent)
                self._add_upstream(upstream, parent)

    def _is_collider(self, u: Node, v: Node, w: Node) -> bool:
        """return True if u-v-w is a collider"""
        if v in self._children[u] and v in self._children[w]:
            return True
        elif v in self._children[u] and v in self._spouses[w]:
            return True
        elif v in self._spouses[u] and v in self._children[w]:
            return True
        elif v in self._spouses[u] and v in self._spouses[w]:
            return True
        else:
            return False

    def _no_other_path(self, i: Node, j: Node, ancestor_dict: dict) -> bool:
        """
        Check if there is any path from i to j other than possibly the direct edge i->j.
        """
        other_ancestors_j = ancestor_dict[j] - {i}
        return (other_ancestors_j & self._children[i]) == set()

    def legitimate_mark_changes(self, verbose=False, strict=True):
        """
        Return directed edges that can be changed to bidirected edges, and bidirected edges that can be changed to
        directed edges.

        Parameters
        ----------
        verbose:
            If True, print each possible mark change and which condition it fails, if any.
        strict:
            If True, check discriminating path condition. Otherwise, check only equality of parents and spouses.

        Return
        ------
        (mark_changes_dir, mark_changes_bidir)
            Directed edges that can be changed to bidirected edges, and bidirected edges that can be changed to directed
            edges (which will be the new directed edge).

        Example
        -------
        >>> g = cd.AncestralGraph(directed={(0, 1)}, bidirected={(1, 2)})
        >>> g.legitimate_mark_changes()
        ({(0, 1)}, {(2, 1)})
        """
        if self._undirected:
            raise ValueError('Only defined for DMAGs')

        if not strict:
            # print("TODO: CHECK")
            ancestor_dict = self.ancestor_dict()
            mark_changes_dir = {
                (i, j) for i, j in self._directed
                if self._parents[i] - self._parents[j] == set() and
                   self._spouses[i] - self._parents[j] - self._spouses[j] == set()
                   and self._no_other_path(i, j, ancestor_dict)
            }
            bidirected = [tuple(e) for e in self._bidirected]
            bidirected_reversed = [tuple(reversed(e)) for e in self._bidirected]
            mark_changes_bidir = {
                (i, j) for i, j in bidirected + bidirected_reversed
                if self._parents[i] - self._parents[j] == set() and
                                   self._spouses[i] - {j} - self._parents[j] - self._spouses[j] == set()
            }
            return mark_changes_dir, mark_changes_bidir

        if strict:
            disc_paths = self.discriminating_paths()
            ancestor_dict = self.ancestor_dict()

            mark_changes_dir = set()
            for i, j in self._directed:
                if verbose: print(f'{i}->{j} => {i}<->{j} ?')
                parents_condition = self._parents[i] - self._parents[j]
                if parents_condition != set():
                    if verbose: print(f'Failed parents condition on {parents_condition}')
                    continue
                spouses_condition = self._spouses[i] - self._spouses[j] - self._parents[j]
                if spouses_condition != set():
                    if verbose: print(f'Failed spouses condition on {spouses_condition}')
                    continue
                ancestral_condition = self._no_other_path(i, j, ancestor_dict)
                # ancestral_condition2 = i not in self.ancestors_of(j, exclude_arcs={(i, j)})
                # print(ancestral_condition == ancestral_condition2)
                # if ancestral_condition != ancestral_condition2:
                #     print(self, i, j, (ancestor_dict[j] - {i}) & self._children[i], ancestor_dict[j], self._children[i])
                if not ancestral_condition:
                    if verbose: print(f'Failed ancestral condition')
                    continue

                # SECOND CONDITION
                disc_paths_for_i = [path for path in disc_paths.keys() if path[-2] == i]
                disc_paths_condition = next((path for path in disc_paths_for_i if path[-1] == j),
                                            None) if disc_paths_for_i else None
                if disc_paths_condition is not None:
                    if verbose: print(f'Failed discriminating path condition on {disc_paths_condition}')
                    continue

                if verbose: print('Passed')
                mark_changes_dir.add((i, j))

            mark_changes_bidir = set()
            forward_edges = {(i, j) for i, j in self._bidirected}
            for i, j in forward_edges | set(map(reversed, forward_edges)):
                if verbose: print(f'{i}<->{j} => {i}->{j} ?')
                parents_condition = self._parents[i] - self._parents[j]
                if parents_condition != set():
                    if verbose: print(f'Failed parents condition on {parents_condition}')
                    continue
                spouses_condition = self._spouses[i] - {j} - self._spouses[j] - self._parents[j]
                if spouses_condition != set():
                    if verbose: print(f'Failed spouses condition on {spouses_condition}')
                    continue
                ancestral_condition = self._no_other_path(i, j, ancestor_dict)

                if not ancestral_condition:
                    if verbose: print('failed ancestral condition')
                    continue

                # SECOND CONDITION
                disc_paths_for_i = [path for path in disc_paths.keys() if path[-2] == i]
                disc_paths_condition = next((path for path in disc_paths_for_i if path[-1] == j),
                                            None) if disc_paths_for_i else None
                if disc_paths_condition is not None:
                    if verbose: print(f'Failed discriminating path condition on {disc_paths_condition}')
                    continue

                if verbose: print('Passed')
                mark_changes_bidir.add((i, j))

            return mark_changes_dir, mark_changes_bidir

    def msep(self, A: Set[Node], B: Set[Node], C: Set[Node]=set()) -> bool:
        """
        Check whether A and B are m-separated given C, using the Bayes ball algorithm.

        Parameters
        ----------
        A:
            Set
        B:
            Set
        C:
            Set

        See Also
        --------
        msep_from_given

        Examples
        --------
        TODO
        """
        # type coercion
        A = core_utils.to_set(A)
        B = core_utils.to_set(B)
        C = core_utils.to_set(C)

        # shade ancestors of C
        shaded_nodes = set(C)
        for node in C:
            self._add_upstream(shaded_nodes, node)

        visited = set()
        # marks whether the node has been encountered along a path where it has a tail or an arrowhead
        _t = 'tail'  # tail
        _a = 'arrowhead'  # arrowhead

        schedule = {(node, _t) for node in A}
        while schedule:
            node, _dir = schedule.pop()
            if node in B: return False
            if (node, _dir) in visited: continue
            visited.add((node, _dir))
            # print(node, _dir)

            # if coming through a tail, won't encounter v-structure
            if _dir == _t and node not in C:
                schedule.update({(parent, _t) for parent in self._parents[node]})
                schedule.update({(child, _a) for child in self._children[node]})
                schedule.update({(spouse, _a) for spouse in self._spouses[node]})
                schedule.update({(nbr, _t) for nbr in self._neighbors[node]})

            if _dir == _a:
                # if coming through an arrowhead and see shaded node, can go through v-structure
                if node in shaded_nodes:
                    schedule.update({(parent, _t) for parent in self._parents[node]})
                    schedule.update({(spouse, _a) for spouse in self._spouses[node]})

                # if coming through an arrowhead and see unconditioned node, can go through children and neighbors
                if node not in C:
                    schedule.update({(child, _a) for child in self._children[node]})
                    schedule.update({(nbr, _a) for nbr in self._neighbors[node]})

        return True

    def msep_from_given(self, A: Set[Node], C: Set[Node]=set()) -> Set[Node]:
        """
        Find all nodes m-separated from A given C.

        Uses algorithm similar to that in Geiger, D., Verma, T., & Pearl, J. (1990).
        Identifying independence in Bayesian networks. Networks, 20(5), 507-534.

        Parameters
        ----------
        A:
            Set
        B:
            Set

        See Also
        --------
        msep

        Examples
        --------
        TODO
        """
        A = core_utils.to_set(A)
        C = core_utils.to_set(C)

        determined = set()
        descendants = set()

        for c in C:
            determined.add(c)
            descendants.add(c)
            self._add_upstream(descendants, c)

        reachable = set()
        i_links = set()
        labeled_links = set()

        for a in A:
            i_links.add((None, a))
            reachable.add(a)

        while True:
            i_p_1_links = set()
            # Find all unlabled links v->w adjacent to at least one link u->v labeled i, such that (u->v,v->w) is a legal pair.
            for link in i_links:
                u, v = link
                for w in self._adjacent[v]:
                    if not u == w and (v, w) not in labeled_links:
                        if self._is_collider(u, v, w):  # Is collider?
                            if v in descendants:
                                i_p_1_links.add((v, w))
                                reachable.add(w)
                        else:  # Not collider
                            if v not in determined:
                                i_p_1_links.add((v, w))
                                reachable.add(w)

            if len(i_p_1_links) == 0:
                break

            labeled_links = labeled_links.union(i_links)
            i_links = i_p_1_links

        return self._nodes.difference(A).difference(C).difference(reachable)


if __name__ == '__main__':
    g = AncestralGraph(nodes=set(range(1, 5)), directed={(1, 2), (2, 4), (3, 2), (3, 4)})
    disc_paths = g.discriminating_paths()



from collections import defaultdict
from copy import deepcopy
import numpy as np
# from cvxopt import spmatrix
from networkx import Graph
from causaldag.classes.custom_types import Node, UndirectedEdge
from typing import Set, Iterable, Dict


class UndirectedGraph:
    def __init__(self, nodes=set(), edges=set()):
        self._nodes = set(nodes)
        self._edges = {frozenset({i, j}) for i, j in edges}
        self._neighbors = defaultdict(set)
        self._degrees = defaultdict(int)
        for i, j in self._edges:
            self._nodes.add(i)
            self._nodes.add(j)
            self._neighbors[i].add(j)
            self._neighbors[j].add(i)
            self._degrees[i] += 1
            self._degrees[j] += 1

    def __eq__(self, other):
        return self._nodes == other._nodes and self._edges == other._edges

    def to_nx(self) -> Graph:
        nx_graph = Graph()
        nx_graph.add_nodes_from(self._nodes)
        nx_graph.add_edges_from(self._edges)
        return nx_graph

    def to_amat(self, node_list=None, sparse=False) -> np.ndarray:
        """
        Return an adjacency matrix for this undirected graph.

        Parameters
        ----------
        node_list:
            List indexing the rows/columns of the matrix.

        Return
        ------
        amat

        Examples
        --------
        TODO
        """
        if not node_list:
            node_list = sorted(self._nodes)
        node2ix = {node: i for i, node in enumerate(node_list)}

        if sparse:
            raise NotImplementedError
            # js, ks = [], []
            # for j, k in self._edges:
            #     js.append(j)
            #     ks.append(k)
            #     js.append(k)
            #     ks.append(j)
            # return spmatrix(1, js, ks)
        amat = np.zeros([self.num_nodes, self.num_nodes], dtype=int)

        for i, j in self._edges:
            amat[node2ix[i], node2ix[j]] = True
            amat[node2ix[j], node2ix[i]] = True
        return amat

    @classmethod
    def from_amat(cls, amat: np.ndarray):
        """
        Return an undirected graph with edges given by amat, i.e. i-j if amat[i,j] != 0

        Parameters
        ----------
        amat:
            Numpy matrix representing edges in the undirected graph.

        Examples
        --------
        >>> amat = np.array([[0, 0, 1], [0, 1, 0], [0, 0, 0]])
        >>> cd.UndirectedGraph.from_amat(amat)
        TODO
        """
        edges = {(i, j) for (i, j), val in np.ndenumerate(amat) if val != 0}
        return UndirectedGraph(nodes=set(range(amat.shape[0])), edges=edges)

    @classmethod
    def from_nx(cls, g: Graph):
        return UndirectedGraph(nodes=g.nodes, edges=g.edges)

    def copy(self, new=True):
        """
        Return a copy of this undirected graph.
        """
        return UndirectedGraph(self._nodes, self._edges)

    @property
    def num_nodes(self) -> int:
        return len(self._nodes)

    @property
    def num_edges(self) -> int:
        return len(self._edges)

    @property
    def degrees(self) -> Dict[Node, int]:
        return {node: self._degrees[node] for node in self._nodes}

    @property
    def neighbors(self) -> Dict[Node, Set[Node]]:
        return {node: self._neighbors[node] for node in self._nodes}

    @property
    def edges(self) -> Set[UndirectedEdge]:
        return self._edges.copy()

    @property
    def nodes(self) -> Set[Node]:
        return self._nodes.copy()

    @property
    def skeleton(self) -> Set[UndirectedEdge]:
        return self.edges

    def has_edge(self, i: Node, j: Node) -> bool:
        """
        Check if the undirected graph has an edge.

        Parameters
        ----------
        i:
            first endpoint of edge.
        j:
            second endpoint of edge.

        Examples
        --------
        TODO
        """
        return frozenset({i, j}) in self._edges

    def neighbors_of(self, node: Node) -> Set[Node]:
        """
        Return the neighbors of a node.

        Parameters
        ----------
        node

        Examples
        --------
        TODO
        """
        return self._neighbors[node].copy()

    # === MUTATORS ===
    def add_edge(self, i: Node, j: Node):
        """
        Add an edge.

        Parameters
        ----------
        i:
            first endpoint of edge to be added.
        j:
            second endpoint of edge to be added.

        See Also
        --------
        add_edges_from
        delete_edge

        Examples
        --------
        TODO
        """
        if frozenset({i, j}) not in self._edges:
            self._edges.add(frozenset({i, j}))
            self._neighbors[i].add(j)
            self._neighbors[j].add(i)
            self._degrees[i] += 1
            self._degrees[j] += 1

    def add_edges_from(self, edges: Iterable):
        """
        Add a set of edges.

        Parameters
        ----------
        edges:
            Edges to be added.

        See Also
        --------
        add_edge
        delete_edges_from

        Examples
        --------
        TODO
        """
        for i, j in edges:
            self.add_edge(i, j)

    def delete_edges_from(self, edges: Iterable):
        """
        Delete a set of edges.

        Parameters
        ----------
        edges:
            Edges to be deleted.

        See Also
        --------
        add_edges_from
        delete_edge

        Examples
        --------
        TODO
        """
        for i, j in edges:
            self.delete_edge(i, j)

    def delete_edge(self, i: Node, j: Node):
        """
        Delete an edge.

        Parameters
        ----------
        i:
            first endpoint of edge to be deleted.
        j:
            second endpoint of edge to be deleted.

        See Also
        --------
        add_edge
        delete_edges_from

        Examples
        --------
        TODO
        """
        self._edges.remove(frozenset({i, j}))
        self._neighbors[i].remove(j)
        self._neighbors[j].remove(i)
        self._degrees[i] -= 1
        self._degrees[j] -= 1

    def delete_node(self, node: Node):
        """
        Delete a node.

        Parameters
        ----------
        node:
            Node to be deleted.

        Examples
        --------
        TODO
        """
        self._nodes.remove(node)
        for j in self._neighbors[node]:
            self._neighbors[j].remove(node)
            self._degrees[j] -= 1
            self._edges.remove(frozenset({node, j}))
        self._neighbors.pop(node, None)
        self._degrees.pop(node, None)

    def to_dag(self, perm):
        raise NotImplementedError





from line_profiler import LineProfiler
import causaldag as cd
from causaldag.inference.structural import pcalg, skeleton
import numpy as np
from causaldag.utils.ci_tests import MemoizedCI_Tester, gauss_ci_suffstat, gauss_ci_test
import random
np.random.seed(1729)
random.seed(1729)

nnodes = 20
nodes = set(range(nnodes))
g = cd.rand.rand_weights(cd.rand.directed_erdos(nnodes, 3/(nnodes-1), 1))
iv_node = random
nsamples = 1000
samples = g.sample(nsamples)
suffstat = gauss_ci_suffstat(samples)
profiler = LineProfiler()


def run_pc():
    for i in range(100):
        ci_tester = MemoizedCI_Tester(gauss_ci_test, suffstat)
        pcalg(nodes, ci_tester, max_cond_set=None, verbose=True)


profiler.add_function(pcalg)
profiler.runcall(run_pc)
profiler.print_stats()



from typing import NewType, Dict, Any, List, Union, Optional
import numpy as np
from dataclasses import dataclass
from scipy.stats import norm


class PerfectInterventionalDistribution:
    """Base class for perfect interventional distributions
    """
    def sample(self, size: int) -> np.ndarray:
        raise NotImplementedError

    def pdf(self, vals: np.ndarray) -> float:
        raise NotImplementedError


class SoftInterventionalDistribution:
    def sample(self, parent_values: np.ndarray, dag, node) -> np.ndarray:
        raise NotImplementedError

    def pdf(self, vals: np.ndarray, parent_values: np.ndarray, dag, node) -> float:
        raise NotImplementedError


InterventionalDistribution = NewType('InterventionalDistribution', Union[PerfectInterventionalDistribution, SoftInterventionalDistribution])
Intervention = NewType('Intervention', Dict[Any, InterventionalDistribution])
PerfectIntervention = NewType('PerfectIntervention', Dict[Any, PerfectInterventionalDistribution])
SoftIntervention = NewType('SoftIntervention', Dict[Any, SoftInterventionalDistribution])


@dataclass
class ScalingIntervention(SoftInterventionalDistribution):
    factor: float = 1
    noise_factor: float = 1
    mean: Optional[float] = None

    def sample(self, parent_values: Optional[np.ndarray], dag, node) -> np.ndarray:
        nsamples, nparents = parent_values.shape
        node_ix = dag._node2ix[node]
        mean = dag._means[node_ix] if self.mean is None else self.mean
        std = dag._variances[node_ix]**.5 * self.noise_factor
        noise = np.random.normal(loc=mean, scale=std, size=nsamples)
        parent_ixs = [dag._node2ix[p] for p in dag._parents[node]]
        if nparents != 0:
            return np.sum(parent_values * dag._weight_mat[parent_ixs, node]*self.factor, axis=1) + noise
        else:
            return noise

    def pdf(self, vals: np.ndarray, parent_values: np.ndarray, dag, node) -> float:
        pass


@dataclass
class ShiftIntervention(SoftInterventionalDistribution):
    shift: float

    def sample(self, parent_values: Optional[np.ndarray], dag, node) -> np.ndarray:
        from causaldag import GaussDAG, SampleDAG
        if isinstance(dag, GaussDAG):
            nsamples, nparents = parent_values.shape
            node_ix = dag._node2ix[node]
            noise = np.random.normal(loc=dag._means[node_ix] + self.shift, scale=dag._variances[node_ix]**.5, size=nsamples)
            parent_ixs = [dag._node2ix[p] for p in dag._parents[node]]
            if nparents != 0:
                return np.sum(parent_values * dag._weight_mat[parent_ixs, node], axis=1) + noise
            else:
                return noise
        elif isinstance(dag, SampleDAG):
            nsamples = parent_values.shape[0]
            samples = np.zeros(nsamples)
            for sample_num in range(nsamples):
                samples[sample_num] = dag.conditionals[node](parent_values[sample_num, :]) + self.shift
            return samples

    def pdf(self, vals: np.ndarray, parent_values: np.ndarray, dag, node) -> float:
        pass


@dataclass
class GaussIntervention(PerfectInterventionalDistribution):
    mean: float = 0
    variance: float = 1

    def sample(self, size: int) -> np.ndarray:
        samples = np.random.normal(loc=self.mean, scale=self.variance**.5, size=size)
        return samples

    def pdf(self, vals: np.ndarray) -> float:
        return norm.pdf(vals, loc=self.mean, scale=self.variance**.5)

    def logpdf(self, vals: np.ndarray) -> float:
        return norm.logpdf(vals, loc=self.mean, scale=self.variance**.5)


@dataclass
class BinaryIntervention(PerfectInterventionalDistribution):
    intervention1: PerfectInterventionalDistribution
    intervention2: PerfectInterventionalDistribution
    p: float = .5

    def sample(self, size: int) -> np.ndarray:
        choices = np.random.binomial(1, self.p, size=size)
        ixs_iv1 = np.where(choices == 1)[0]
        ixs_iv2 = np.where(choices == 0)[0]
        samples = np.zeros(size)
        samples[ixs_iv1] = self.intervention1.sample(len(ixs_iv1))
        samples[ixs_iv2] = self.intervention2.sample(len(ixs_iv2))
        return samples

    def pdf(self, vals: np.ndarray) -> float:
        return self.p * self.intervention1.pdf(vals) + (1 - self.p) * self.intervention2.pdf(vals)


@dataclass
class MultinomialIntervention(PerfectInterventionalDistribution):
    pvals: List[float]
    interventions: List[PerfectInterventionalDistribution]

    def sample(self, size: int) -> np.ndarray:
        choices = np.random.choice(list(range(len(self.interventions))), size=size, p=self.pvals)
        samples = np.zeros(size)
        for ix, iv in enumerate(self.interventions):
            ixs_iv = np.where(choices == ix)[0]
            samples[ixs_iv] = iv.sample(len(ixs_iv))
        return samples

    def pdf(self, vals: np.ndarray) -> float:
        raise NotImplementedError


@dataclass
class ConstantIntervention(PerfectInterventionalDistribution):
    val: float

    def sample(self, size: int) -> np.ndarray:
        return np.ones(size) * self.val

    def pdf(self, vals: np.ndarray) -> float:
        return np.all(vals == self.val).astype(float)




from causaldag.classes import UndirectedGraph
import itertools as itr
import numpy as np
from numpy.linalg import inv
from numpy import sqrt, diag, eye
from causaldag.utils import core_utils
from numpy.random import multivariate_normal


class GGM(UndirectedGraph):
    def __init__(self, precision_matrix: np.ndarray, means: np.ndarray=None):
        nnodes = precision_matrix.shape[0]
        self.precision_matrix = precision_matrix
        self.means = means if means is not None else np.ones(nnodes)
        nodes = set(range(nnodes))
        edges = {(i, j) for i, j in itr.combinations(nodes, 2) if precision_matrix[i, j] != 0}
        super().__init__(set(nodes), edges)

        self._covariance = None
        self._correlation = None

    @classmethod
    def from_covariance(cls, covariance: np.ndarray, means: np.ndarray=None):
        g = GGM(inv(covariance), means)
        g._covariance = covariance
        return g

    @classmethod
    def from_adjacency(cls, adjacency_matrix: np.ndarray, variances: np.ndarray=None, means: np.ndarray=None):
        nnodes = adjacency_matrix.shape[0]
        variances = variances if variances is not None else np.ones(nnodes)
        means = means if means is not None else np.zeros(nnodes)

        id_ = eye(nnodes)
        a = adjacency_matrix
        if (variances == 1).all():
            precision = (id_ - a) @ (id_ - a).T
        else:
            precision = (id_ - a) @ diag(variances ** -1) @ (id_ - a).T

        return GGM(precision, means)

    def _ensure_covariance(self):
        if self._covariance is None:
            self._covariance = inv(self.precision_matrix)

    def _ensure_correlation(self):
        if self._correlation is None:
            S = self.covariance
            self._correlation = S / sqrt(diag(S)) / sqrt(diag(S)).reshape([-1, 1])

    @property
    def covariance(self):
        self._ensure_covariance()
        return self._covariance

    @property
    def correlation(self):
        self._ensure_correlation()
        return self._correlation

    def partial_correlation(self, i, j, cond_set):
        """
        Return the partial correlation of i and j conditioned on `cond_set`.

        Parameters
        ----------
        i:
        j:
        cond_set:

        Examples
        --------
        TODO
        """
        cond_set = core_utils.to_set(cond_set)
        if len(cond_set) == 0:
            return self.correlation[i, j]
        else:
            theta = inv(self.correlation[np.ix_([i, j, *cond_set], [i, j, *cond_set])])
            return -theta[0, 1] / np.sqrt(theta[0, 0] * theta[1, 1])

    def sample(self, nsamples: int=1) -> np.ndarray:
        return multivariate_normal(self.means, self.covariance, nsamples)

    def to_gauss_dag(self, perm):
        """
        Return a GaussDAG with the same mean and covariance as this GGM, and is a minimal IMAP of this GGM
        consistent with the node ordering `perm`.

        Parameters
        ----------
        perm:
            The desired permutation, or total order, of the nodes in the result.

        Returns
        -------

        Examples
        --------
        TODO
        """
        from causaldag import DAG, GaussDAG

        d = DAG(nodes=self.nodes)
        ixs = list(itr.chain.from_iterable(((f, s) for f in range(s)) for s in range(len(perm))))
        for i, j in ixs:
            pi_i, pi_j = perm[i], perm[j]
            if not np.isclose(self.partial_correlation(pi_i, pi_j, d.markov_blanket(pi_i)), 0):
                d.add_arc(pi_i, pi_j, unsafe=True)

        arcs = dict()
        means = []
        Sigma = self.covariance
        variances = []
        for i in perm:
            ps = list(d.parents_of(i))

            # === LINEAR REGRESSION TO FIND EDGE WEIGHTS
            S_xx = Sigma[np.ix_(ps, ps)]
            S_xy = Sigma[ps, i]
            coeffs = inv(S_xx) @ S_xy

            # === COMPUTE MEAN AND VARIANCE
            mean = self.means[i] - self.means[ps] @ coeffs.T
            variance = Sigma[i, i] - Sigma[i, ps] @ coeffs

            for p, coeff in zip(ps, coeffs):
                print(p, i)
                arcs[(p, i)] = coeff
            means.append(mean)
            variances.append(variance)

        return GaussDAG(list(range(self.num_nodes)), arcs, means=means, variances=variances)


if __name__ == '__main__':
    P = np.eye(3)
    P[0, 1] = P[1, 0] = .1
    P[2, 1] = P[1, 2] = .1
    g = GGM(P)
    samples = g.sample(1000)
    cov_sample = np.cov(samples, rowvar=False)

    from causaldag import GaussDAG
    A = np.array([[0, .1, 0], [0, 0, .1], [0, 0, 0]])
    gdag = GaussDAG.from_amat(A)
    ggm = GGM.from_covariance(gdag.covariance)
    gdag2 = ggm.to_gauss_dag([0, 1, 2])
    print(gdag.weight_mat)
    print(gdag2.weight_mat)





import numpy as np
from causaldag import DAG, GaussDAG, SampleDAG
import itertools as itr
from typing import Union, List, Callable
from networkx import barabasi_albert_graph


def _coin(p, size=1):
    return np.random.binomial(1, p, size=size)


def unif_away_zero(low=.25, high=1, size=1, all_positive=False):
    if all_positive:
        return np.random.uniform(low, high, size=size)
    return (_coin(.5, size) - .5) * 2 * np.random.uniform(low, high, size=size)


def directed_erdos(nnodes, density, size=1, as_list=False) -> Union[DAG, List[DAG]]:
    """
    Generate random Erdos-Renyi DAG(s) on `nnodes` nodes with density `density`.

    Parameters
    ----------
    nnodes:
        Number of nodes in each graph.
    density:
        Probability of any edge.
    size:
        Number of graphs.
    as_list:
        If True, always return as a list, even if only one DAG is generated.

    Examples
    --------
    >>> d = cd.rand.directed_erdos(5, .5)
    """
    if size == 1:
        bools = _coin(density, size=int(nnodes * (nnodes - 1) / 2))
        arcs = {(i, j) for (i, j), b in zip(itr.combinations(range(nnodes), 2), bools) if b}
        d = DAG(nodes=set(range(nnodes)), arcs=arcs)
        return [d] if as_list else d
    else:
        return [directed_erdos(nnodes, density) for _ in range(size)]


def rand_weights(dag, rand_weight_fn=unif_away_zero) -> GaussDAG:
    """
    Generate a GaussDAG from a DAG, with random edge weights independently drawn from `rand_weight_fn`.

    Parameters
    ----------
    dag:
        DAG
    rand_weight_fn:
        Function to generate random weights.

    Examples
    --------
    >>> d = cd.DAG(arcs={(1, 2), (2, 3)})
    >>> g = cd.rand.rand_weights(d)
    """
    weights = rand_weight_fn(size=len(dag.arcs))
    return GaussDAG(nodes=list(range(len(dag.nodes))), arcs=dict(zip(dag.arcs, weights)))


def rand_nn_functions(dag: DAG, num_layers=3) -> SampleDAG:
    s = SampleDAG(dag._nodes, arcs=dag._arcs)
    for node in dag._nodes:
        def conditional(parent_vals):
            p = len(parent_vals)
            vals = parent_vals
            for _ in range(num_layers):
                a = np.random.random((p, p))*2
                vals = a @ vals
                vals = np.where(vals > 0, vals, vals*.01)
            return np.random.random(p)*2 @ vals + np.random.laplace(0, 1)
        s.set_conditional(node, conditional)
    return s


def directed_random_graph(nnodes: int, random_graph_model: Callable, size=1, as_list=False) -> Union[DAG, List[DAG]]:
    if size == 1:
        # generate a random undirected graph
        edges = random_graph_model(nnodes).edges

        # generate a random permutation
        random_permutation = np.arange(nnodes)
        np.random.shuffle(random_permutation)

        arcs = []
        for edge in edges:
            node1, node2 = edge
            node1_position = np.where(random_permutation == node1)[0][0]
            node2_position = np.where(random_permutation == node2)[0][0]
            if node1_position < node2_position:
                source = node1
                endpoint = node2
            else:
                source = node2
                endpoint = node1
            arcs.append((source, endpoint))
        d = DAG(nodes=set(range(nnodes)), arcs=arcs)
        return [d] if as_list else d
    else:
        return [directed_random_graph(nnodes, random_graph_model) for _ in range(size)]


def directed_barabasi(nnodes: int, nattach: int, size=1, as_list=False) -> Union[DAG, List[DAG]]:
    random_graph_model = lambda nnodes: barabasi_albert_graph(nnodes, nattach)
    return directed_random_graph(nnodes, random_graph_model, size=size, as_list=as_list)


__all__ = [
    'directed_erdos',
    'rand_weights',
    'unif_away_zero',
    'directed_barabasi',
    'directed_random_graph',
    'rand_nn_functions'
]





import numpy as np


class PAG:
    def __init__(self, nodes, adjacencies, arrowheads, tails):
        self._nodes = nodes.copy()
        self._adjacencies = adjacencies.copy()
        self._arrowheads = arrowheads.copy()
        self._tails = tails.copy()

    def __eq__(self, other):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def copy(self):
        raise NotImplementedError

    @property
    def nnodes(self):
        return len(self._nodes)

    @classmethod
    def from_amat(cls, amat):
        p = amat.shape[0]
        adjacencies = set()
        arrowheads = set()
        tails = set()
        for (i, j), val in np.ndenumerate(amat):
            if val == 1:
                adjacencies.add(frozenset({i, j}))
            if val == 2:
                arrowheads.add((i, j))
                adjacencies.add(frozenset({i, j}))
            if val == 3:
                tails.add((i, j))
                adjacencies.add(frozenset({i, j}))
        return PAG(set(range(p)), adjacencies, arrowheads, tails)

    def to_amat(self):
        # TODO: check
        amat = np.zeros([self.nnodes, self.nnodes])
        for i, j in self._adjacencies:
            amat[i, j] = 1
            amat[j, i] = 1
        for i, j in self._tails:
            amat[i, j] = 3
        for i, j in self._arrowheads:
            amat[i, j] = 2
        return amat

    def add_adjacency(self, i, j):
        """Add the adjacency (i, j)
        """
        pass

    def add_arrowhead(self, i, j):
        """Add an arrowhead at j to the adjacency (i, j)
        """
        pass

    def add_tail(self, i, j):
        """Add a tail at i to the adjacency (i, j)
        """
        pass

    @property
    def nodes(self):
        return set(self._nodes)

    @property
    def adjacencies(self):
        return set(self._adjacencies)

    @property
    def arrowheads(self):
        return set(self._arrowheads)

    @property
    def tails(self):
        return set(self._tails)

    @property
    def circles(self):
        raise NotImplementedError

    @property
    def skeleton(self):
        return self._adjacencies.copy()

    def shd_skeleton(self, other):
        return len(self.skeleton.symmetric_difference(other.skeleton))




"""
===============================
Estimation of differences between directed graphs given two datasets representing two settings.
===============================
This module contains functions for estimating the differences between two causal 
directed acyclic graph (DAG) models given samples from each model.

References
----------
    [1] Wang, Y., Squires, C., Belyaeva, A., & Uhler, C. (2018). Direct estimation of differences in causal graphs. 
    In Advances in Neural Information Processing Systems (pp. 3770-3781).
    [2] Meinshausen, N. and Buhlmann, P. (2010). Stability selection.
    Journal of the Royal Statistical Society: Series B (Statistical Methodology), 72(4), pp.417-473.
"""

from causaldag.structure_learning.difference.difference_ug import dci_undirected_graph
from causaldag.utils.ci_tests import gauss_ci_suffstat
from causaldag.utils.core_utils import powerset
from causaldag.utils.regression import RegressionHelper
from scipy.special import ncfdtr
from numpy.linalg import inv
import numpy as np
import itertools
from joblib import Parallel, delayed
from sklearn.utils import safe_mask
from sklearn.utils.random import sample_without_replacement


def dci(
        X1,
        X2,
        alpha_ug: float = 1.0,
        alpha_skeleton: float = 0.1,
        alpha_orient: float = 0.1,
        max_set_size: int = None,
        difference_ug: list = None,
        max_iter: int = 1000,
        edge_threshold: float = 0.05,
        verbose: int = 0
):
    """
    Uses the Difference Causal Inference (DCI) algorithm to estimate the difference-DAG between two settings.

    Parameters
    ----------
    X1: array, shape = [n_samples, n_features]
        First dataset.    
    X2: array, shape = [n_samples, n_features]
        Second dataset.
    alpha_ug: float, default = 1.0
        L1 regularization parameter for estimating the difference undirected graph via KLIEP algorithm.
    alpha_skeleton: float, default = 0.1
        Significance level parameter for determining presence of edges in the skeleton of the difference graph. 
        Lower alpha_skeleton results in sparser difference graph.
    alpha_orient: float, default = 0.1
        Significance level parameter for determining orientation of an edge. 
        Lower alpha_orient results in more directed edges in the difference-DAG.
    max_set_size: int, default = None
        Maximum conditioning set size used to test regression invariance.
        Smaller maximum conditioning set size results in faster computation time. For large datasets recommended max_set_size is 3.
    difference_ug: list, default = None
        List of tuples that represents edges in the difference undirected graph. If difference_ug is None, 
        KLIEP algorithm for estimating the difference undirected graph will be run. 
        If the number of nodes is small, difference_ug could be taken to be the complete graph between all the nodes.
    max_iter: int, default = 1000
        Maximum number of iterations for gradient descent in KLIEP algorithm.
    edge_threshold: float, default = 0.05
        Edge weight cutoff for keeping an edge for KLIEP algorithm (all edges above or equal to this threshold are kept).
    verbose: int, default = 0
        The verbosity level of logging messages.

    See Also
    --------
    dci_undirected_graph, dci_skeleton, dci_orient

    Returns
    -------
    adjacency_matrix: array, shape  = [n_features, n_features]
        Estimated difference-DAG. Edges that were found to be different between two settings but the orientation
        could not be determined, are represented by assigning 1 in both directions, i.e. adjacency_matrix[i,j] = 1
        and adjacency_matrix[j,i] = 1. Otherwise for oriented edges, only adjacency_matrix[i,j] = 1 is assigned. 
        Assignment of 0 in the adjacency matrix represents no edge.

    References
    ----------
        [1] Wang, Y., Squires, C., Belyaeva, A., & Uhler, C. (2018). Direct estimation of differences in causal graphs. 
        In Advances in Neural Information Processing Systems (pp. 3770-3781).
    """

    assert 0 <= alpha_skeleton <= 1, "alpha_skeleton must be in [0,1] range."
    assert 0 <= alpha_orient <= 1, "alpha_orient must be in [0,1] range."

    num_nodes = X1.shape[1]
    # obtain sufficient statistics
    suffstat1 = gauss_ci_suffstat(X1)
    suffstat2 = gauss_ci_suffstat(X2)
    rh1 = RegressionHelper(suffstat1)
    rh2 = RegressionHelper(suffstat2)

    # compute the difference undirected graph via KLIEP if the differece_ug is not provided
    if difference_ug is None:
        difference_ug = dci_undirected_graph(X1, X2, alpha=alpha_ug, max_iter=max_iter, edge_threshold=edge_threshold,
                                             verbose=verbose)

    # estimate the skeleton of the difference-DAG 
    skeleton = dci_skeleton(difference_ug, rh1, rh2, alpha=alpha_skeleton, max_set_size=max_set_size, verbose=verbose)
    # orient edges of the skeleton of the difference-DAG
    edges_oriented, edges_unoriented = dci_orient(skeleton, rh1, rh2, alpha=alpha_orient, max_set_size=max_set_size,
                                                  verbose=verbose)

    adjacency_matrix = edges2adjacency(num_nodes, edges_unoriented, undirected=True) + edges2adjacency(num_nodes,
                                                                                                       edges_oriented,
                                                                                                       undirected=False)
    return adjacency_matrix


def dci_stability_selection(
        X1,
        X2,
        alpha_ug_grid: list = [0.1, 1, 10],
        alpha_skeleton_grid: list = [0.1, 0.5],
        alpha_orient_grid: list = [0.001, 0.1],
        max_set_size: int = None,
        difference_ug: list = None,
        max_iter: int = 1000,
        edge_threshold: float = 0.05,
        sample_fraction: float = 0.7,
        n_bootstrap_iterations: int = 50,
        bootstrap_threshold: float = 0.5,
        n_jobs: int = 1,
        random_state: int = None,
        verbose: int = 0
):
    """
    Runs Difference Causal Inference (DCI) algorithm with stability selection to estimate the difference-DAG between two settings. 
    Bootstrap samples are generated from two input datasets and DCI is run across bootstrap samples and across different 
    combinations of hyperparameters. Edges that reliably appear across different runs are considered stable and are output as the difference-DAG.

    Parameters
    ----------
    X1: array, shape = [n_samples, n_features]
        First dataset.    
    X2: array, shape = [n_samples, n_features]
        Second dataset.
    alpha_ug_grid: array-like, default = [0.1, 1, 10]
        Grid of values to iterate over representing L1 regularization parameter for estimating the difference undirected graph via KLIEP algorithm.
    alpha_skeleton: array-like, default = [0.1, 0.5]
        Grid of values to iterate over representing significance level parameter for determining presence of edges in the skeleton of the difference graph. 
        Lower alpha_skeleton results in sparser difference graph.
    alpha_orient: array-like, default = [0.001, 0.1]
        Grid of values to iterate over representing significance level parameter for determining orientation of an edge. 
        Lower alpha_orient results in more directed edges in the difference-DAG.
    max_set_size: int, default = None
        Maximum conditioning set size used to test regression invariance.
        Smaller maximum conditioning set size results in faster computation time. For large datasets recommended max_set_size is 3.
    difference_ug: list, default = None
        List of tuples that represents edges in the difference undirected graph. If difference_ug is None, 
        KLIEP algorithm for estimating the difference undirected graph will be run. 
        If the number of nodes is small, difference_ug could be taken to be the complete graph between all the nodes.
    max_iter: int, default = 1000
        Maximum number of iterations for gradient descent in KLIEP algorithm.
    edge_threshold: float, default = 0.05
        Edge weight cutoff for keeping an edge for KLIEP algorithm (all edges above or equal to this threshold are kept).
    sample_fraction: float, default = 0.7
        The fraction of samples to be used in each bootstrap sample.
        Should be between 0 and 1. If 1, all samples are used.
    n_bootstrap_iterations: int, default = 50
        Number of bootstrap samples to create.
    bootstrap_threshold: float, default = 0.5
        Threshold defining the minimum cutoff value for the stability scores. Edges with stability scores above
        the bootstrap_threshold are kept as part of the difference-DAG.
    n_jobs: int, default = 1
        Number of jobs to run in parallel.
    random_state: int, default = None
        Seed used by the random number generator.
    verbose: int, default = 0
        The verbosity level of logging messages.

    See Also
    --------
    dci, dci_undirected_graph, dci_skeleton, dci_orient

    Returns
    -------
    adjacency_matrix: array, shape  = [n_features, n_features]
        Estimated difference-DAG. Edges that were found to be different between two settings but the orientation
        could not be determined, are represented by assigning 1 in both directions, i.e. adjacency_matrix[i,j] = 1
        and adjacency_matrix[j,i] = 1. Otherwise for oriented edges, only adjacency_matrix[i,j] = 1 is assigned. 
        Assignment of 0 in the adjacency matrix represents no edge.
    stability_scores: array, shape = [n_params, n_features, n_features]
        Stability score of each edge for for each combination of hyperparameters.

    References
    ----------
        [1] Wang, Y., Squires, C., Belyaeva, A., & Uhler, C. (2018). Direct estimation of differences in causal graphs. 
        In Advances in Neural Information Processing Systems (pp. 3770-3781).
        [2] Meinshausen, N. and Buhlmann, P. (2010). Stability selection.
           Journal of the Royal Statistical Society: Series B (Statistical Methodology), 72(4), pp.417-473.
    """

    _, n_variables = X1.shape
    n_params = len(alpha_ug_grid) * len(alpha_skeleton_grid) * len(alpha_orient_grid)

    hyperparams = itertools.product(alpha_ug_grid, alpha_skeleton_grid, alpha_orient_grid)
    stability_scores = np.zeros((n_params, n_variables, n_variables))

    for idx, params in enumerate(hyperparams):
        if verbose > 0:
            print(
                "Fitting estimator for alpha_ug = %.5f, alpha_skeleton = %.5f, alpha_orient = %.5f with %d bootstrap iterations" %
                (params[0], params[1], params[2], n_bootstrap_iterations))

        bootstrap_samples1 = bootstrap_generator(n_bootstrap_iterations, sample_fraction,
                                                 X1, random_state=random_state)
        bootstrap_samples2 = bootstrap_generator(n_bootstrap_iterations, sample_fraction,
                                                 X2, random_state=random_state)

        bootstrap_results = Parallel(n_jobs, verbose=verbose
                                     )(delayed(dci)(X1[safe_mask(X1, subsample1), :],
                                                    X2[safe_mask(X2, subsample2), :],
                                                    alpha_ug=params[0],
                                                    alpha_skeleton=params[1],
                                                    alpha_orient=params[2],
                                                    max_set_size=max_set_size,
                                                    difference_ug=difference_ug,
                                                    max_iter=max_iter,
                                                    edge_threshold=edge_threshold,
                                                    verbose=verbose)
                                       for subsample1, subsample2 in zip(bootstrap_samples1, bootstrap_samples2))

        stability_scores[idx] = np.array(bootstrap_results).mean(axis=0)

    adjacency_matrix = choose_stable_variables(stability_scores, bootstrap_threshold=bootstrap_threshold)
    return adjacency_matrix, stability_scores


def choose_stable_variables(stability_scores, bootstrap_threshold=0.5):
    """Returns adjacency matrix corresponding to edges with stability scores above threshold."""
    return (stability_scores.max(axis=0) > bootstrap_threshold).astype('float')


def bootstrap_generator(n_bootstrap_iterations, sample_fraction, X, random_state=None):
    """Generates bootstrap samples from dataset."""
    n_samples = len(X)
    n_subsamples = np.floor(sample_fraction * n_samples).astype(int)
    for _ in range(n_bootstrap_iterations):
        subsample = sample_without_replacement(n_samples, n_subsamples, random_state=random_state)
        yield subsample


def dci_skeleton(
        difference_ug: list,
        rh1: RegressionHelper,
        rh2: RegressionHelper,
        alpha: float = 0.1,
        max_set_size: int = 3,
        verbose: int = 0
):
    """
    Estimates the skeleton of the difference-DAG.

    Parameters
    ----------
    difference_ug: list
        List of tuples that represents edges in the difference undirected graph.
    rh1: RegressionHelper
        Sufficient statistics estimated based on samples in the first dataset, stored in RegressionHelper class.
    rh2: RegressionHelper
        Sufficient statistics estimated based on samples in the second dataset, stored in RegressionHelper class.
    alpha: float, default = 0.1
        Significance level parameter for determining presence of edges in the skeleton of the difference graph.
        Lower alpha results in sparser difference graph.
    max_set_size: int, default = None
        Maximum conditioning set size used to test regression invariance.
        Smaller maximum conditioning set size results in faster computation time. For large datasets recommended max_set_size is 3.
    verbose: int, default = 0
        The verbosity level of logging messages.

    See Also
    --------
    dci, dci_undirected_graph, dci_orient

    Returns
    -------
    skeleton: set
        Set of edges in the skeleton of the difference-DAG.
    """

    if verbose > 0:
        print("DCI skeleton estimation...")

    assert 0 <= alpha <= 1, "alpha must be in [0,1] range."

    n1 = rh1.suffstat['n']
    n2 = rh2.suffstat['n']
    nodes = get_nodes_in_graph(difference_ug)
    skeleton = {frozenset({i, j}) for i, j in difference_ug}

    for i, j in difference_ug:
        for cond_set in powerset(nodes - {i, j}, r_max=max_set_size):
            cond_set_i, cond_set_j = [*cond_set, j], [*cond_set, i]

            # calculate regression coefficients (j regressed on cond_set_j) for both datasets
            beta1_i, var1_i, precision1 = rh1.regression(i, cond_set_i)
            beta2_i, var2_i, precision2 = rh2.regression(i, cond_set_i)

            # compute statistic and p-value
            j_ix = cond_set_i.index(j)
            stat_i = (beta1_i[j_ix] - beta2_i[j_ix]) ** 2 * \
                     inv(var1_i * precision1 / (n1 - 1) + var2_i * precision2 / (n2 - 1))[j_ix, j_ix]
            pval_i = ncfdtr(1, n1 + n2 - len(cond_set_i) - len(cond_set_j), 0, stat_i)
            pval_i = 2 * min(pval_i, 1 - pval_i)

            #  remove i-j from skeleton if i regressed on (j, cond_set) is invariant
            i_invariant = pval_i > alpha
            if i_invariant:
                if verbose > 0:
                    print("Removing edge %d-%d since p-value=%.5f < alpha=%.5f" % (i, j, pval_i, alpha))
                skeleton.remove(frozenset({i, j}))
                break

            # calculate regression coefficients (i regressed on cond_set_i) for both datasets
            beta1_j, var1_j, precision1 = rh1.regression(j, cond_set_j)
            beta2_j, var2_j, precision2 = rh2.regression(j, cond_set_j)

            # compute statistic and p-value
            i_ix = cond_set_j.index(i)
            stat_j = (beta1_j[i_ix] - beta2_j[i_ix]) ** 2 * \
                     inv(var1_j * precision1 / (n1 - 1) + var2_j * precision2 / (n2 - 1))[i_ix, i_ix]
            pval_j = 1 - ncfdtr(1, n1 + n2 - len(cond_set_i) - len(cond_set_j), 0, stat_j)
            pval_j = 2 * min(pval_j, 1 - pval_j)

            #  remove i-j from skeleton if j regressed on (i, cond_set) is invariant
            j_invariant = pval_j > alpha
            if j_invariant:
                if verbose > 0:
                    print("Removing edge %d-%d since p-value=%.5f < alpha=%.5f" % (i, j, pval_j, alpha))
                skeleton.remove(frozenset({i, j}))
                break

    return skeleton


def dci_orient(
        skeleton: set,
        rh1: RegressionHelper,
        rh2: RegressionHelper,
        alpha: float = 0.1,
        max_set_size: int = 3,
        verbose: int = 0
):
    """
    Orients edges in the skeleton of the difference DAG.

    Parameters
    ----------
    skeleton: set
        Set of edges in the skeleton of the difference-DAG.
    rh1: RegressionHelper
        Sufficient statistics estimated based on samples in the first dataset, stored in RegressionHelper class.
    rh2: RegressionHelper
        Sufficient statistics estimated based on samples in the second dataset, stored in RegressionHelper class.
    alpha: float, default = 0.1
        Significance level parameter for determining orientation of an edge.
        Lower alpha results in more directed edges in the difference-DAG.
    max_set_size: int, default = 3
        Maximum conditioning set size used to test regression invariance.
        Smaller maximum conditioning set size results in faster computation time. For large datasets recommended max_set_size is 3.
    verbose: int, default = 0
        The verbosity level of logging messages.

    See Also
    --------
    dci, dci_undirected_graph, dci_skeleton

    Returns
    -------
    oriented_edges: set
        Set of edges in the skeleton of the difference-DAG for which directionality could be determined.
    unoriented_edges: set
        Set of edges in the skeleton of the difference-DAG for which directionality could not be determined.
    """

    if verbose > 0:
        print("DCI edge orientation...")

    assert 0 <= alpha <= 1, "alpha must be in [0,1] range."

    nodes = {i for i, j in skeleton} | {j for i, j in skeleton}
    oriented_edges = set()

    n1 = rh1.suffstat['n']
    n2 = rh2.suffstat['n']
    for i, j in skeleton:
        for cond_i, cond_j in zip(powerset(nodes - {i}, r_max=max_set_size), powerset(nodes - {j}, r_max=max_set_size)):
            # compute residual variances for i
            beta1_i, var1_i, _ = rh1.regression(i, cond_i)
            beta2_i, var2_i, _ = rh2.regression(i, cond_i)
            # compute p-value for invariance of residual variances for i
            pvalue_i = ncfdtr(n1 - len(cond_i), n2 - len(cond_i), 0, var1_i / var2_i)
            pvalue_i = 2 * min(pvalue_i, 1 - pvalue_i)

            # compute residual variances for j
            beta1_j, var1_j, _ = rh1.regression(j, cond_j)
            beta2_j, var2_j, _ = rh2.regression(j, cond_j)
            # compute p-value for invariance of residual variances for j
            pvalue_j = ncfdtr(n1 - len(cond_j), n2 - len(cond_j), 0, var1_j / var2_j)
            pvalue_j = 2 * min(pvalue_j, 1 - pvalue_j)

            if ((pvalue_i > alpha) | (pvalue_j > alpha)):
                # orient the edge according to highest p-value
                if pvalue_i > pvalue_j:
                    edge = (j, i) if j in cond_i else (i, j)
                else:
                    edge = (i, j) if i in cond_j else (j, i)
                oriented_edges.add(edge)

                if verbose > 0:
                    print("Oriented (%d, %d) as %s" % (i, j, edge))
                break

    unoriented_edges = skeleton - {frozenset({i, j}) for i, j in oriented_edges}
    return oriented_edges, unoriented_edges


def edges2adjacency(num_nodes, edge_set, undirected=False):
    """
    Returns adjacency_matrix given a set of edges. If the edges are considered undirected,
    then the adjacency matrix will be symmetric.

    Parameters
    ----------
    num_nodes: int
        Number of nodes in the graph.
    edge_set: set
        Set of edges in the graph.    
    undirected: bool, default = False
        Whether to consider the edges in the edge set as directed or undirected.

    Returns
    -------
    adjacency_matrix: array, shape  = [num_nodes, num_nodes]
        Adjacency matrix.
    """

    adjacency_matrix = np.zeros((num_nodes, num_nodes))
    for parent, child in edge_set:
        adjacency_matrix[parent, child] = 1
        if undirected:
            adjacency_matrix[child, parent] = 1
    return adjacency_matrix


def get_nodes_in_graph(graph):
    """
    Returns nodes that are in the graph.
    """
    return set(np.unique(graph))



import itertools as itr
from causaldag.classes import UndirectedGraph
from causaldag.utils.ci_tests import CI_Tester, gauss_ci_test
from numpy import sqrt, log1p, ndenumerate, errstate
from scipy.special import erf


def threshold_ug(nodes: set, ci_tester: CI_Tester) -> UndirectedGraph:
    """
    Estimate an undirected graph by testing whether each pair of nodes is independent given all others.

    Parameters
    ----------
    nodes:
        Nodes in the graph.
    ci_tester:
        Conditional independence tester.

    Examples
    --------
    TODO
    """
    if hasattr(ci_tester, 'ci_test') and ci_tester.ci_test == gauss_ci_test:
        return threshold_ug_gauss(ci_tester)
    edges = {(i, j) for i, j in itr.combinations(nodes, 2) if not ci_tester.is_ci(i, j, nodes - {i, j})}
    return UndirectedGraph(nodes, edges)


def threshold_ug_gauss(ci_tester):
    """
    Estimate an undirected graph by testing whether each pair of nodes is independent given all others,
    which reduces to thresholding partial correlations (after the Fisher z-transform) for multivariate Gaussian
    data.

    Parameters
    ----------
    ci_tester:
        Conditional independence tester.

    Examples
    --------
    TODO
    """
    rho = ci_tester.suffstat.get('rho')
    if rho is not None:
        r = rho
    else:
        raise ValueError

    n = ci_tester.suffstat['n']
    p = ci_tester.suffstat['C'].shape[0]
    n_cond = p-2

    # note: log1p(2r/(1-r)) = log((1+r)/(1-r)) but is more numerically stable for r near 0
    # r = 1 causes warnings but gives the correct answer
    with errstate(divide='ignore', invalid='ignore'):
        statistic = sqrt(n - n_cond - 3) * abs(.5 * log1p(2*r/(1 - r)))

    p_values = 1 - .5*(1 + erf(statistic/sqrt(2)))
    alpha = ci_tester.kwargs.get('alpha')
    alpha = alpha if alpha is not None else 1e-5
    edges = {(i, j) for (i, j), p in ndenumerate(p_values) if p < alpha if i != j}

    return UndirectedGraph(set(range(p)), edges)





import causaldag as cd
from causaldag.structure_learning.dag import perm2dag, min_degree_alg_amat, gsp
from causaldag.structure_learning.undirected import threshold_ug
import random
import itertools as itr


def apply_lmc(imap, i, j):
    if imap.has_directed(i, j):
        imap.remove_directed(i, j)
        imap.add_bidirected(i, j)
    else:
        imap.remove_bidirected(i, j)
        imap.add_directed(i, j)


def get_lmc_altered_edges(imap: cd.AncestralGraph, i, j, ci_tester):
    """
    Given an IMAP and a legitimate mark change applied to it, test which edges can be removed after the legitimate
    mark change.
    """
    imap_copy = imap.copy()
    apply_lmc(imap_copy, i, j)

    desc_j = imap.descendants_of(j) | {j}

    for k, l in list(imap.directed) + list(imap.bidirected):
        if k in desc_j or l in desc_j:
            c = imap_copy.ancestors_of({k, l}) - {k, l}
            if ci_tester.is_ci(k, l, c):
                imap_copy.remove_edge(k, l)
    removed_dir = imap.directed - imap_copy.directed
    if imap.has_directed(i, j): removed_dir = removed_dir - {(i, j)}
    removed_bidir = bidirected_frozenset(imap) - bidirected_frozenset(imap_copy)
    if imap.has_bidirected(i, j): removed_bidir = removed_bidir - {frozenset({i, j})}

    return removed_dir, removed_bidir


def bidirected_frozenset(m):
    return frozenset({frozenset({*e}) for e in m._bidirected})


def gspo(
        nodes: set,
        ci_tester,
        depth=4,
        initial_imap='permutation',
        strict=True,
        verbose=False,
        max_iters=float('inf'),
        nruns=5,
):
    """
    Estimate a MAG using the Greedy Sparsest Poset algorithm.

    Parameters
    ----------
    nodes:
        Labels of nodes in the graph.
    ci_tester:
        A conditional independence tester, which has a method is_ci taking two sets A and B, and a conditioning set C,
        and returns True/False.
    depth:
        Maximum depth in depth-first search. Use None for infinite search depth.
    initial_imap:
        String indicating how to obtain the initial IMAP. Must be "permutation" or "empty".
    strict:
        If True, check discriminating paths condition for legitimate mark changes.
    verbose:
        If True, print information about algorithm progress.
    max_iters:
        Maximum number of depth-first search steps without score improvement before stopping.
    nruns:
        Number of times to run the algorithm (each run may vary due to randomness in tie-breaking and/or starting
        imap.

    Return
    ------
    An estimated MAG
    """
    if initial_imap == 'permutation':
        ug = threshold_ug(nodes, ci_tester)
        amat = ug.to_amat()
        perms = [min_degree_alg_amat(amat) for _ in range(nruns)]
        dags = [perm2dag(perm, ci_tester) for perm in perms]
        starting_imaps = [cd.AncestralGraph(dag.nodes, directed=dag.arcs) for dag in dags]
    elif initial_imap == 'empty':
        edges = {(i, j) for i, j in itr.combinations(nodes, 2) if not ci_tester.is_ci(i, j)}
        starting_imaps = [cd.AncestralGraph(nodes, bidirected=edges) for _ in range(nruns)]
    elif initial_imap == 'gsp':
        ug = threshold_ug(nodes, ci_tester)
        amat = ug.to_amat()
        perms = [min_degree_alg_amat(amat) for _ in range(nruns)]
        dags = [gsp(nodes, ci_tester, nruns=1, initial_permutations=[perm]) for perm in perms]
        starting_imaps = [cd.AncestralGraph(dag.nodes, directed=dag.arcs) for dag, _ in dags]

    get_alt_edges = get_lmc_altered_edges

    sparsest_imap = None
    for r in range(nruns):
        current_imap = starting_imaps[r]
        if verbose: print(f"Starting run {r} with {current_imap.num_edges} edges")

        # TODO: BOTTLENECK
        current_lmcs_directed, current_lmcs_bidirected = current_imap.legitimate_mark_changes(strict=strict)
        current_lmcs = current_lmcs_directed | current_lmcs_bidirected

        # TODO: BOTTLENECK
        lmcs2altered_edges = [
            (lmc, get_alt_edges(current_imap, *lmc, ci_tester))
            for lmc in current_lmcs
        ]
        lmcs2altered_edges = [(lmc, (a, b)) for lmc, (a, b) in lmcs2altered_edges if a is not None]
        lmcs2edge_delta = [
            (lmc, len(removed_dir) + len(removed_bidir))
            for lmc, (removed_dir, removed_bidir) in lmcs2altered_edges
        ]

        mag2number = dict()
        graph_counter = 0
        trace = []
        iters_since_improvement = 0
        while True:
            if iters_since_improvement > max_iters:
                break

            mag_hash = (frozenset(current_imap._directed), bidirected_frozenset(current_imap))
            if mag_hash not in mag2number:
                mag2number[mag_hash] = graph_counter
            graph_num = mag2number[mag_hash]
            if verbose: print(f"Number of visited MAGs: {len(mag2number)}. Exploring MAG #{graph_num} with {current_imap.num_edges} edges.")
            max_delta = max([delta for lmc, delta in lmcs2edge_delta], default=0)

            sparser_exists = max_delta > 0
            keep_searching_mec = len(trace) != depth and len(lmcs2altered_edges) > 0

            if sparser_exists:
                trace = []

                lmc_ix = random.choice([ix for ix, (lmc, delta) in enumerate(lmcs2edge_delta) if delta == max_delta])
                (i, j), (removed_dir, removed_bidir) = lmcs2altered_edges.pop(lmc_ix)
                apply_lmc(current_imap, i, j)
                current_imap.remove_edges(removed_dir | removed_bidir)

                if verbose: print(f"Starting over at a sparser IMAP with {current_imap.num_edges} edges")
            elif keep_searching_mec:
                if verbose: print(f"{'='*len(trace)}Continuing search through the MEC at {current_imap.num_edges} edges. "
                                  f"Picking from {len(lmcs2altered_edges)} neighbors of #{graph_num}.")
                trace.append((current_imap.copy(), current_lmcs, lmcs2altered_edges, lmcs2edge_delta))
                (i, j), _ = lmcs2altered_edges.pop(0)
                lmcs2edge_delta.pop(0)
                apply_lmc(current_imap, i, j)
            elif len(trace) != 0:  # BACKTRACK IF POSSIBLE
                if verbose: print(f"{'='*len(trace)}Backtracking")
                current_imap, current_lmcs, lmcs2altered_edges, lmcs2edge_delta = trace.pop()
                iters_since_improvement += 1
            else:
                break

            # IF WE MOVED TO A NOVEL IMAP, WE NEED TO UPDATE LMCs
            if sparser_exists or keep_searching_mec:
                graph_counter += 1
                current_lmcs_dir, current_lmcs_bidir = current_imap.legitimate_mark_changes(strict=strict)
                current_lmcs = current_lmcs_dir | current_lmcs_bidir
                lmcs2altered_edges = [
                    (lmc, get_alt_edges(current_imap, *lmc, ci_tester))
                    for lmc in current_lmcs
                ]
                lmcs2altered_edges = [(lmc, (a, b)) for lmc, (a, b) in lmcs2altered_edges if a is not None]
                current_directed, current_bidirected = frozenset(current_imap.directed), bidirected_frozenset(current_imap)

                # === FILTER OUT ALREADY-VISITED IMAPS
                filtered_lmcs2altered_edges = []
                for lmc, (removed_dir, removed_bidir) in lmcs2altered_edges:
                    if current_imap.has_directed(*lmc):
                        new_directed = current_directed - {lmc} - removed_dir
                        new_bidirected = current_bidirected | {frozenset({*lmc})} - {frozenset({*e}) for e in removed_bidir}
                    else:
                        new_directed = current_directed | {lmc} - removed_dir
                        new_bidirected = current_bidirected - {frozenset({*lmc})} - {frozenset({*e}) for e in removed_bidir}

                    if (new_directed, new_bidirected) not in mag2number:
                        filtered_lmcs2altered_edges.append((lmc, (removed_dir, removed_bidir)))
                lmcs2altered_edges = filtered_lmcs2altered_edges

                lmcs2edge_delta = [
                    (lmc, len(removed_dir)+len(removed_bidir))
                    for lmc, (removed_dir, removed_bidir) in lmcs2altered_edges
                ]
        if sparsest_imap is None or sparsest_imap.num_edges > current_imap.num_edges:
            sparsest_imap = current_imap

    return current_imap



from typing import Union, List, Dict


def dsep_invariance_test(
        suffstat: Dict,
        context,
        i,
        cond_set: Union[List[int], int]=None
):
    dag = suffstat['dag']
    intervened_nodes = suffstat[context]
    return dict(reject=not dag.is_invariant(i, intervened_nodes, cond_set=cond_set))



from typing import Any, Union, List
import numpy as np
from causaldag.utils.ci_tests import kernels
from scipy.stats import gamma
from causaldag.utils.ci_tests._utils import residuals
from causaldag.utils.core_utils import to_list


def hsic_test_vector(x: np.ndarray, y: np.ndarray, sig: float=1/np.sqrt(2), alpha=0.05):
    if x.ndim == 1:
        x = x.reshape((len(x), 1))
    if y.ndim == 1:
        y = y.reshape((len(y), 1))
    n = x.shape[0]
    if y.shape[0] != n:
        raise ValueError("Y should have the same number of samples as X")

    n = x.shape[0]
    H = np.eye(n) - np.ones([n, n])/n
    kernel_precision = 1/(sig**2)

    # === COMPUTE CENTRALIZED KERNEL MATRICES
    kx = kernels.rbf_kernel(x, kernel_precision)
    ky = kernels.rbf_kernel(y, kernel_precision)
    kx_centered = kernels.center(kx)
    ky_centered = kernels.center(ky)

    # === COMPUTE STATISTIC
    statistic = 1/n**2 * np.sum(kx_centered * ky_centered.T)  # SAME AS trace(kx_centered @ ky_centered)

    mu_x = 1/(n*(n-1)) * np.sum(kx - np.diag(np.diag(kx)))  # SUM OFF-DIAGONALS
    mu_y = 1/(n*(n-1)) * np.sum(ky - np.diag(np.diag(ky)))
    mean_approx = 1/n * (1 + mu_x*mu_y - mu_x - mu_y)
    var_approx = 2*(n-4)*(n-5)/(n*(n-1)*(n-2)*(n-3)) * np.sum(kx_centered * kx_centered.T) * np.sum(ky_centered * ky_centered.T) / n**4

    k_approx = mean_approx ** 2 / var_approx
    prec_approx = var_approx / mean_approx

    critval = gamma.ppf(1-alpha, k_approx, scale=prec_approx)
    p_value = 1 - gamma.cdf(statistic, k_approx, scale=prec_approx)

    return dict(
        statistic=statistic,
        critval=critval,
        p_value=p_value,
        reject=statistic > critval,
        mean_approx=mean_approx,
        var_approx=var_approx
    )


def hsic_test(
        suffstat: Any,
        i: int,
        j: int,
        cond_set: Union[List[int], int]=None,
        alpha: float=0.05
):
    cond_set = to_list(cond_set)
    if len(cond_set) == 0:
        return hsic_test_vector(suffstat[:, i], suffstat[:, j], alpha=alpha)
    else:
        residuals_i, residuals_j = residuals(suffstat, i, j, cond_set)
        return hsic_test_vector(residuals_i, residuals_j, alpha=alpha)









from numpy import sqrt, log1p, ndenumerate, errstate
from scipy.special import erf


def covariance_graph_gauss(ci_tester):
    """
    Estimate a covariance graph by testing whether each pair of nodes is independent, which reduces to
    thresholding correlations (after the Fisher z-transform) for multivariate Gaussian data.

    Parameters
    ----------
    ci_tester:
        Conditional independence tester.

    Examples
    --------
    TODO
    """
    corr = ci_tester.suffstat['C']

    n = ci_tester.suffstat['n']

    # note: log1p(2r/(1-r)) = log((1+r)/(1-r)) but is more numerically stable for r near 0
    # r = 1 causes warnings but gives the correct answer
    with errstate(divide='ignore', invalid='ignore'):
        statistic = sqrt(n - 3) * abs(.5 * log1p(2*corr/(1 - corr)))

    p_values = 1 - .5*(1 + erf(statistic/sqrt(2)))
    alpha = ci_tester.kwargs.get('alpha')
    alpha = alpha if alpha is not None else 1e-5
    edges = {(i, j) for (i, j), p in ndenumerate(p_values) if p < alpha if i != j}





from causaldag import PDAG
from causaldag import UndirectedGraph
from causaldag.utils.ci_tests import CI_Tester
import itertools as itr


def skeleton(nodes: set, ci_tester: CI_Tester, max_cond_set: int=None, verbose=False):
    """
    Estimate the skeleton of an underlying DAG using the order-independent skeleton estimation method of
    Colombo and Maathius (2014).

    Parameters
    ----------
    nodes:
        Labels of nodes in the graph.
    ci_tester:
        A conditional independence tester, which has a method is_ci taking two sets A and B, and a conditioning set C,
        and returns True/False.
    max_cond_set:
        Maximum size of conditioning set tested to separate nodes.
    verbose:
        If True, print edges as they are removed, along with the separating set responsible for removing them.

    See Also
    --------
    pcalg

    Returns
    -------
    (skeleton, sepset)
    """
    nnodes = len(nodes)
    ug = UndirectedGraph(edges=set(itr.combinations(nodes, 2)))
    sepset = {}
    max_cond_set = max_cond_set if max_cond_set is not None else nnodes-2
    for c_size in range(max_cond_set+1):
        adjacencies = ug.neighbors
        for i, j in itr.permutations(nodes, 2):
            if ug.has_edge(i, j) and len(adjacencies[i] - {j}) >= c_size:
                for cond_set in itr.combinations(adjacencies[i] - {j}, c_size):
                    if ci_tester.is_ci(i, j, cond_set):
                        if verbose: print(f"Removing {i}-{j}, separated by {cond_set}")
                        ug.delete_edge(i, j)
                        sepset[frozenset({i, j})] = cond_set
                        break
    return ug, sepset


def pcalg(
        nodes,
        ci_tester: CI_Tester=None,
        skel=None,
        sepset=None,
        solve_conflict: bool=False,
        max_cond_set: int=None,
        verbose: bool=False
):
    """
    Use the PC (Peters-Clark) algorithm to estimate the Markov equivalence class of the data-generating DAG.

    Parameters
    ----------
    nodes:
        Labels of nodes in the graph.
    ci_tester:
        A conditional independence tester, which has a method is_ci taking two sets A and B, and a conditioning set C,
        and returns True/False.
    skel:
        An estimated skeleton. If not provided, uses the `skeleton` method to estimate.
    sepset:
        The separating sets for non-adjacent nodes in the estimated skeleton.
    solve_conflict:
        If False, any disagreements on v-structures are simply overwritten. If True, allow both orientations
        (represented by a bidirected edge).
    verbose:
        If True, print decisions made by the algorithm.

    See Also
    --------
    gsp

    Returns
    -------
    est_dag
    """
    if solve_conflict:
        raise NotImplementedError
    if ci_tester is None:
        if skel is None or sepset is None:
            raise ValueError("Must provide either ci_tester or skeleton and sepset dictionary")
    if ci_tester is not None:
        skel, sepset = skeleton(nodes, ci_tester, max_cond_set=max_cond_set, verbose=verbose)
    adjacencies = skel.neighbors

    arcs = set()
    for i, k in itr.combinations(nodes, 2):
        if not skel.has_edge(i, k):
            for j in adjacencies[i] & adjacencies[k]:
                if j not in sepset[frozenset({i, k})]:
                    if not solve_conflict:
                        arcs.discard((j, k))
                        arcs.discard((j, k))
                    arcs.add((i, j))
                    arcs.add((k, j))

    cpdag = PDAG(nodes=nodes, arcs=arcs, edges=skel.edges-{frozenset({*arc}) for arc in arcs})
    cpdag.to_complete_pdag(verbose=verbose, solve_conflict=solve_conflict)

    return cpdag


if __name__ == '__main__':
    import causaldag as cd
    from causaldag.utils.ci_tests import MemoizedCI_Tester, dsep_test

    import numpy as np
    import random
    np.random.seed(9890142)
    random.seed(9890142)

    nnodes = 15
    ngraphs = 10
    exp_nbrs = 3
    dags = cd.rand.directed_erdos(nnodes, exp_nbrs/(nnodes-1), ngraphs)
    nodes = set(range(nnodes))
    ci_testers = [MemoizedCI_Tester(dsep_test, d) for d in dags]
    est_skels = [skeleton(set(range(nnodes)), ci_tester)[0] for ci_tester in ci_testers]
    false_positives = [est_skel.edges - d.skeleton for est_skel, d in zip(est_skels, dags)]
    false_negatives = [d.skeleton - est_skel.edges for est_skel, d in zip(est_skels, dags)]
    print(false_positives)
    print(false_negatives)

    est_cpdags = [pcalg(nodes, ci_tester) for ci_tester in ci_testers]
    true_cpdags = [d.cpdag() for d in dags]
    correct_cpdag = [true_cpdag == est_cpdag for true_cpdag, est_cpdag in zip(true_cpdags, est_cpdags)]
    shds = [true_cpdag.shd(est_cpdag) for true_cpdag, est_cpdag in zip(true_cpdags, est_cpdags)]
    print(correct_cpdag)

